from shexer.io.shex.formater.statement_serializers.frequency_strategy.abs_freq_serializer import AbsFreqSerializer
from shexer.io.shex.formater.statement_serializers.frequency_strategy.ratio_freq_serializer import RatioFreqSerializer
from shexer.io.shex.formater.statement_serializers.frequency_strategy.mixed_frequency_strategy import MixedFrequencyStrategy
from shexer.io.shex.formater.statement_serializers.base_statement_serializer import BaseStatementSerializer
from shexer.io.shex.formater.statement_serializers.fixed_prop_choice_statement_serializer import FixedPropChoiceStatementSerializer
from shexer.consts import RATIO_INSTANCES, ABSOLUTE_INSTANCES, MIXED_INSTANCES

class StSerializerFactory(object):
    """
    This factory offers public method with Singletons*. They are not really singletons, as a battery
    of objects are always initialized, no matter which calls the factory receives.

    But the point here is that there is only one isntance of each type of serializer.

    """

    def __init__(self, freq_mode, decimals, instantiation_property_str, disable_comments, frequency_property,
                 namespaces_dict, comments_to_annotations, extra_info_property, absolute_count_property):
        self._freq_serializer = self._build_freq_serializer(freq_mode=freq_mode,
                                                            decimals=decimals,
                                                            freq_prop=frequency_property,
                                                            abs_prop=absolute_count_property,
                                                            namespaces_dict=namespaces_dict)

        self._direct_base = BaseStatementSerializer(
                instantiation_property_str=instantiation_property_str,
                disable_comments=disable_comments,
                is_inverse=False,
                frequency_serializer=self._freq_serializer,
                frequency_property=frequency_property,
                namespaces_dict=namespaces_dict,
                comments_to_annotations=comments_to_annotations,
                extra_info_prop=extra_info_property,
                absolute_count_prop=absolute_count_property)
        self._inverse_base = BaseStatementSerializer(
                instantiation_property_str=instantiation_property_str,
                disable_comments=disable_comments,
                is_inverse=True,
                frequency_serializer=self._freq_serializer,
                frequency_property=frequency_property,
                namespaces_dict=namespaces_dict,
                comments_to_annotations=comments_to_annotations,
                extra_info_prop=extra_info_property,
                absolute_count_prop=absolute_count_property)
        self._direct_choice = FixedPropChoiceStatementSerializer(
                instantiation_property_str=instantiation_property_str,
                disable_comments=disable_comments,
                is_inverse=False,
                frequency_serializer=self._freq_serializer,
                frequency_property=frequency_property,
                namespaces_dict=namespaces_dict,
                comments_to_annotations=comments_to_annotations,
                extra_info_prop=extra_info_property,
                absolute_count_prop=absolute_count_property)
        self._inverse_choice = FixedPropChoiceStatementSerializer(
                instantiation_property_str=instantiation_property_str,
                disable_comments=disable_comments,
                is_inverse=True,
                frequency_serializer=self._freq_serializer,
                frequency_property=frequency_property,
                namespaces_dict=namespaces_dict,
                comments_to_annotations=comments_to_annotations,
                extra_info_prop=extra_info_property,
                absolute_count_prop=absolute_count_property)

    def get_base_serializer(self, is_inverse):
        return self._direct_base if not is_inverse else self._inverse_base

    def get_choice_serializer(self, is_inverse):
        return self._direct_choice if not is_inverse else self._inverse_choice

    def _build_freq_serializer(self, freq_mode, decimals, namespaces_dict, freq_prop, abs_prop):
        if freq_mode == RATIO_INSTANCES:
            return RatioFreqSerializer(decimals=decimals,
                                       namespaces_dict=namespaces_dict,
                                       frequency_ratio_property=freq_prop,
                                       frequency_absolute_property=abs_prop)
        elif freq_mode == ABSOLUTE_INSTANCES:
            return AbsFreqSerializer(namespaces_dict=namespaces_dict,
                                     frequency_ratio_property=freq_prop,
                                     frequency_absolute_property=abs_prop)
        elif freq_mode == MIXED_INSTANCES:
            return MixedFrequencyStrategy(decimals=decimals,
                                          namespaces_dict=namespaces_dict,
                                          frequency_ratio_property=freq_prop,
                                          frequency_absolute_property=abs_prop
                                          )
        else:
            raise ValueError("Unrecognized frequency strategy for serialization. "
                             "Check you used a valid value in the instances_report_mode param")



from shexer.utils.factories.triple_yielders_factory import get_triple_yielder
from shexer.core.profiling.class_profiler import ClassProfiler
from shexer.core.profiling.federated_source_class_profiler import FederatedSourceClassProfiler
from shexer.utils.target_elements import tune_target_classes_if_needed
from shexer.utils.dict import reverse_keys_and_values


def get_class_profiler(target_classes_dict, source_file, list_of_source_files, input_format,
                       instantiation_property_str,
                       namespaces_to_ignore=None,
                       infer_numeric_types_for_untyped_literals=False,
                       raw_graph=None,
                       namespaces_dict=None,
                       url_input=None,
                       list_of_url_input=None,
                       rdflib_graph=None,
                       shape_map_file=None,
                       shape_map_raw=None,
                       track_classes_for_entities_at_last_depth_level=True,
                       depth_for_building_subgraph=1,
                       url_endpoint=None,
                       strict_syntax_with_corners=False,
                       target_classes=None,
                       file_target_classes=None,
                       built_remote_graph=None,
                       built_shape_map=None,
                       remove_empty_shapes=True,
                       limit_remote_instances=-1,
                       inverse_paths=False,
                       all_classes_mode=False,
                       compression_mode=None,
                       disable_endpoint_cache=None,
                       detect_minimal_iri=False,
                       examples_mode=None,
                       federated_sources=None,
                       shapes_namespace=None):
    yielder = get_triple_yielder(source_file=source_file,
                                 list_of_source_files=list_of_source_files,
                                 input_format=input_format,
                                 namespaces_to_ignore=namespaces_to_ignore,
                                 raw_graph=raw_graph,
                                 allow_untyped_numbers=infer_numeric_types_for_untyped_literals,
                                 namespaces_dict=namespaces_dict,
                                 url_input=url_input,
                                 list_of_url_input=list_of_url_input,
                                 rdflib_graph=rdflib_graph,
                                 shape_map_file=shape_map_file,
                                 shape_map_raw=shape_map_raw,
                                 track_classes_for_entities_at_last_depth_level=track_classes_for_entities_at_last_depth_level,
                                 depth_for_building_subgraph=depth_for_building_subgraph,
                                 url_endpoint=url_endpoint,
                                 instantiation_property=instantiation_property_str,
                                 strict_syntax_with_corners=strict_syntax_with_corners,
                                 target_classes=target_classes,
                                 file_target_classes=file_target_classes,
                                 built_remote_graph=built_remote_graph,
                                 built_shape_map=built_shape_map,
                                 limit_remote_instances=limit_remote_instances,
                                 inverse_paths=inverse_paths,
                                 all_classes_mode=all_classes_mode,
                                 compression_mode=compression_mode,
                                 disable_endpoint_cache=disable_endpoint_cache)

    if federated_sources is None or len(federated_sources) == 0:
        return ClassProfiler(triples_yielder=yielder,
                             instances_dict=target_classes_dict,
                             instantiation_property_str=instantiation_property_str,
                             original_target_classes=None
                             if target_classes is None
                             else tune_target_classes_if_needed(
                                 list_target_classes=target_classes,
                                 prefix_namespaces_dict=reverse_keys_and_values(namespaces_dict)),
                             original_shape_map=built_shape_map,
                             remove_empty_shapes=remove_empty_shapes,
                             inverse_paths=inverse_paths,
                             detect_minimal_iri=detect_minimal_iri,
                             examples_mode=examples_mode,
                             namespaces_dict=namespaces_dict,
                             shapes_namespace=shapes_namespace)
    else:
        return FederatedSourceClassProfiler(triples_yielder=yielder,
                                            instances_dict=target_classes_dict,
                                            instantiation_property_str=instantiation_property_str,
                                            original_target_classes=None
                                            if target_classes is None
                                            else tune_target_classes_if_needed(
                                                list_target_classes=target_classes,
                                                prefix_namespaces_dict=reverse_keys_and_values(namespaces_dict)),
                                            original_shape_map=built_shape_map,
                                            remove_empty_shapes=remove_empty_shapes,
                                            inverse_paths=inverse_paths,
                                            detect_minimal_iri=detect_minimal_iri,
                                            examples_mode=examples_mode,
                                            list_of_federated_objects=federated_sources,
                                            namespaces_dict=namespaces_dict,
                                            shapes_namespace=shapes_namespace)



import unittest
from shexer.shaper import Shaper
from test.const import BASE_FILES
from test.t_utils import rdflib_exist_triple, rdflib_counts_triples
import os.path as pth
from shexer.consts import SHACL_TURTLE
from rdflib import Graph

_BASE_DIR = BASE_FILES + "shacl" + pth.sep


class TestShapeMapCompatibility(unittest.TestCase):

    def test_shape_map_compatibility_single_node(self):
        shape_map_raw = "<http://dbpedia.org/resource/Ladilla_Rusa>@<http://example.org/shapes/MusicGroup>"

        shaper = Shaper(
            shape_map_raw=shape_map_raw,
            url_endpoint="https://dbpedia.org/sparql",
            instantiation_property="http://www.w3.org/1999/02/22-rdf-syntax-ns#type",
            namespaces_dict={
                "http://dbpedia.org/resource/": "dbr",
                "http://dbpedia.org/ontology/": "dbo",
                "http://www.w3.org/2001/XMLSchema#": "xsd",
                "http://www.w3.org/1999/02/22-rdf-syntax-ns#": "rdf",
                "http://www.w3.org/2000/01/rdf-schema#": "rdfs"
            },
            disable_comments=True
        )
        str_result = shaper.shex_graph(string_output=True, output_format=SHACL_TURTLE)
        graph = Graph()
        graph.parse(data=str_result)
        self.assertTrue(rdflib_exist_triple(graph=graph,
                                            t_subject="<http://example.org/shapes/MusicGroup>",
                                            t_predicate="<http://www.w3.org/ns/shacl#targetNode>",
                                            t_object="<http://dbpedia.org/resource/Ladilla_Rusa>"))

    def test_shape_map_compatibility_sparql_selector(self):
        shape_map_raw = "SPARQL'select ?s where {?s a <http://dbpedia.org/ontology/MusicalArtist>} LIMIT 2'" \
                        "@<http://example.org/shapes/MusicGroup>"

        shaper = Shaper(
            shape_map_raw=shape_map_raw,
            url_endpoint="https://dbpedia.org/sparql",
            track_classes_for_entities_at_last_depth_level=False,
            depth_for_building_subgraph=1,
            instantiation_property="http://www.w3.org/1999/02/22-rdf-syntax-ns#type",
            namespaces_dict={
                "http://dbpedia.org/resource/": "dbr",
                "http://dbpedia.org/ontology/": "dbo",
                "http://www.w3.org/2001/XMLSchema#": "xsd",
                "http://www.w3.org/1999/02/22-rdf-syntax-ns#": "rdf",
                "http://www.w3.org/2000/01/rdf-schema#": "rdfs"
            },
            disable_comments=True
        )
        str_result = shaper.shex_graph(string_output=True, output_format=SHACL_TURTLE)
        graph = Graph()
        graph.parse(data=str_result)
        self.assertTrue(rdflib_counts_triples(graph=graph,
                                              t_subject="<http://example.org/shapes/MusicGroup>",
                                              t_predicate="<http://www.w3.org/ns/shacl#targetNode>",
                                              t_object=None) == 2)

    def test_shape_map_compatibility_focus_expression(self):
        shape_map_raw = "{dbr:Felipe_VI dbo:predecessor FOCUS}" \
                        "@<http://example.org/shapes/Someone>"
        # We are assuming in this test that this information won't change in DBpedia:
        # Felipe VI has a single predecessor (dbr:Juan_Carlos_I)

        shaper = Shaper(
            shape_map_raw=shape_map_raw,
            url_endpoint="https://dbpedia.org/sparql",
            track_classes_for_entities_at_last_depth_level=False,
            depth_for_building_subgraph=1,
            instantiation_property="http://www.w3.org/1999/02/22-rdf-syntax-ns#type",
            namespaces_dict={
                "http://dbpedia.org/resource/": "dbr",
                "http://dbpedia.org/ontology/": "dbo",
                "http://www.w3.org/2001/XMLSchema#": "xsd",
                "http://www.w3.org/1999/02/22-rdf-syntax-ns#": "rdf",
                "http://www.w3.org/2000/01/rdf-schema#": "rdfs"
            },
            disable_comments=True
        )
        str_result = shaper.shex_graph(string_output=True, output_format=SHACL_TURTLE)
        graph = Graph()
        graph.parse(data=str_result)
        self.assertTrue(rdflib_counts_triples(graph=graph,
                                              t_subject="<http://example.org/shapes/Someone>",
                                              t_predicate="<http://www.w3.org/ns/shacl#targetNode>",
                                              t_object=None) == 1)



import unittest
from shexer.shaper import Shaper
from test.const import G1, BASE_FILES, default_namespaces, G1_ALL_CLASSES_NO_COMMENTS
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth
from shexer.consts import TURTLE

_BASE_DIR = BASE_FILES + "node_types" + pth.sep

class TestNodeTypes(unittest.TestCase):

    def test_proeprty_to_literal_and_iri(self):
        shaper = Shaper(all_classes_mode=True,
                        graph_file_input=_BASE_DIR + "property_to_IRI_and_literal.ttl",
                        namespaces_dict=default_namespaces(),
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "property_to_IRI_and_literal.shex",
                                                      str_target=str_result))




from shexer.model.statement import POSITIVE_CLOSURE, KLEENE_CLOSURE, OPT_CARDINALITY
from shexer.model.const_elem_types import NONLITERAL_ELEM_TYPE, BNODE_ELEM_TYPE, IRI_ELEM_TYPE
from shexer.model.fixed_prop_choice_statement import FixedPropChoiceStatement
from shexer.model.statement import Statement
from shexer.io.shex.formater.statement_serializers.st_serializers_factory import StSerializerFactory
from shexer.core.shexing.strategy.minimal_iri_strategy.annotate_min_iri_strategy import AnnotateMinIriStrategy
from shexer.core.shexing.strategy.minimal_iri_strategy.ignore_min_iri_strategy import IgnoreMinIriStrategy
from shexer.model.shape import STARTING_CHAR_FOR_SHAPE_NAME


_DIRECT_ST_SERIALIZER = 0
_INVERSE_ST_SERIALIZER = 1


class AbstractShexingStrategy(object):

    def __init__(self, class_shexer):
        self._class_shexer = class_shexer
        self._namespaces_dict = class_shexer._namespaces_dict
        self._allow_opt_cardinality = class_shexer._allow_opt_cardinality
        self._disable_comments = self._class_shexer._disable_comments
        self._instantiation_property_str = self._class_shexer._instantiation_property_str
        self._keep_less_specific = self._class_shexer._keep_less_specific
        self._discard_useless_positive_closures = self._class_shexer._discard_useless_positive_closures
        self._tolerance = self._class_shexer._tolerance
        self._disable_or_statements = self._class_shexer._disable_or_statements
        self._all_compliant_mode = self._class_shexer._all_compliant_mode
        self._disable_exact_cardinality = self._class_shexer._disable_exact_cardinality
        self._allow_redundant_or = self._class_shexer._allow_redundant_or
        self._shape_names_dict = self._class_shexer._shape_names_dict

        self._strategy_min_iri = AnnotateMinIriStrategy(class_shexer._class_min_iris_dict) \
            if class_shexer._detect_minimal_iri \
            else IgnoreMinIriStrategy()

        self._statement_serializer_factory = StSerializerFactory(freq_mode=class_shexer._instances_report_mode,
                                                                 decimals=class_shexer._decimals,
                                                                 instantiation_property_str=self._instantiation_property_str,
                                                                 disable_comments=self._disable_comments,
                                                                 frequency_property=self._class_shexer._frequency_property,
                                                                 namespaces_dict=self._namespaces_dict,
                                                                 comments_to_annotations=class_shexer._comments_to_annotations,
                                                                 extra_info_property=self._class_shexer._extra_info_property,
                                                                 absolute_count_property=self._class_shexer._absolute_count_property)


    def yield_base_shapes(self, acceptance_threshold):
        for a_shape in self._yield_base_shapes_direction_aware(acceptance_threshold=acceptance_threshold):
            self._strategy_min_iri.annotate_shape_iri(a_shape)
            yield a_shape

    def _yield_base_shapes_direction_aware(self, acceptance_threshold):
        raise NotImplementedError()

    def set_valid_shape_constraints(self, shape):
        raise NotImplementedError()

    def remove_statements_to_gone_shapes(self, shape, shape_names_to_remove):
        raise NotImplementedError()

    def _tune_list_of_valid_statements(self, valid_statements):
        """
        This method modifies the statements objects received --> no return needed
        :param valid_statements:
        :return:
        """
        if not len(valid_statements) == 0:
            valid_statements.sort(reverse=True, key=lambda x: x.probability)  # Restoring order completely
                                                                              # before changing cardinalities

            if self._all_compliant_mode:
                self._modify_cardinalities_of_statements_non_compliant_with_all_instances(valid_statements)

            if self._disable_exact_cardinality:
                self._generalize_exact_cardinalities(valid_statements)

            if self._disable_comments:
                self._remove_comments_from_statements(valid_statements)

    def _select_valid_statements_of_shape(self, original_statements):
        if len(original_statements) == 0:
            return []

        for a_statement in original_statements:  # TODO Refactor!!! This is not the place to set the serializer
            self._set_serializer_object_for_statements(a_statement)

        result = self._group_constraints_with_same_prop_and_obj(original_statements)
        result = self._group_node_constraints(result)

        return result

    def _compute_frequency(self, number_of_instances, n_ocurrences_statement):
        return float(n_ocurrences_statement) / number_of_instances

    def _modify_cardinalities_of_statements_non_compliant_with_all_instances(self, statements):
        for a_statement in statements:
            if a_statement.probability != 1:
                self._change_statement_cardinality_to_all_compliant(a_statement)

    def _change_statement_cardinality_to_all_compliant(self, statement):
        comment_for_current_sentence = self._turn_statement_into_comment(statement, self._namespaces_dict)
        statement.add_comment(comment=comment_for_current_sentence,
                              insert_first=True)
        statement.cardinality = OPT_CARDINALITY if \
            self._allow_opt_cardinality and statement.cardinality == 1 \
            else KLEENE_CLOSURE
        # statement.probability = 1


    @staticmethod
    def _turn_statement_into_comment(a_statement, namespaces_dict):
        return a_statement.comment_representation(namespaces_dict=namespaces_dict)

    def _generalize_exact_cardinalities(self, statements):
        for a_statement in statements:
            if type(a_statement.cardinality) == int and a_statement.cardinality > 1:
                a_statement.cardinality = POSITIVE_CLOSURE

    def _remove_comments_from_statements(self, valid_statements):
        for a_statement in valid_statements:
            a_statement.remove_comments()

    def _set_serializer_object_for_statements(self, statement):
        statement.serializer_object = self._statement_serializer_factory.get_base_serializer(
            is_inverse=statement.is_inverse
        )

    def _group_constraints_with_same_prop_and_obj(self, candidate_statements):  # TODO REFACTORING
        result = []
        already_visited = set()
        for i in range(0, len(candidate_statements)):
            a_statement = candidate_statements[i]
            if a_statement not in already_visited:
                already_visited.add(a_statement)
                group_to_decide = MergeableConstraints(a_statement)  # TODO
                for j in range(i + 1, len(candidate_statements)):
                    if self._statements_have_same_tokens(a_statement,
                                                         candidate_statements[j]):
                        group_to_decide.add_constraint(candidate_statements[j])
                        already_visited.add(candidate_statements[j])
                if len(group_to_decide) == 1:
                    result.append(a_statement)
                else:
                    result.append(self._decide_best_statement_with_cardinalities_in_comments(group_to_decide))
        return result

    def _statements_have_same_tokens(self, st1, st2):
        if st1.st_property == st2.st_property and st1.st_type == st2.st_type:
            return True
        return False

    def _decide_best_statement_with_cardinalities_in_comments(self, mergeable_constraints):  # TODO REFACTORING
        if self._discard_useless_positive_closures:
            if self._is_a_group_of_statements_with_useless_positive_closure(mergeable_constraints):
                return self._statement_for_a_group_with_a_useless_positive_closure(mergeable_constraints)
        mergeable_constraints.sort()
        result = None
        if self._keep_less_specific:
            for a_statement in mergeable_constraints.constraints():
                if a_statement.cardinality == POSITIVE_CLOSURE:
                    result = a_statement
                    break
            if result is None:
                result = mergeable_constraints.get(0)
        else:
            for a_statement in mergeable_constraints.constraints():
                if a_statement.cardinality != POSITIVE_CLOSURE:
                    result = a_statement
                    break
            if result is None:
                result = mergeable_constraints.get(0)

        for a_statement in mergeable_constraints.constraints():
            if a_statement.cardinality != result.cardinality:
                result.add_comment(self._turn_statement_into_comment(a_statement, self._namespaces_dict))
        return result

    def _is_a_group_of_statements_with_useless_positive_closure(self, list_of_candidate_sentences):  # todo during refactor
        if len(list_of_candidate_sentences) != 2:
            return False
        if abs(list_of_candidate_sentences.get(0).probability - list_of_candidate_sentences.get(1).probability) > self._tolerance:
            return False
        one_if_there_is_a_single_positive_closure = -1
        for a_statement in list_of_candidate_sentences.constraints():
            if POSITIVE_CLOSURE == a_statement.cardinality:
                one_if_there_is_a_single_positive_closure *= -1
        if one_if_there_is_a_single_positive_closure == 1:
            return True
        return False

    def _statement_for_a_group_with_a_useless_positive_closure(self, group_of_candidate_statements):  # TODO DURING REFACTOR
        for a_statement in group_of_candidate_statements.constraints():
            if a_statement.cardinality != POSITIVE_CLOSURE:
                return a_statement
        raise ValueError("The received group does not contain any statement with positive closure")

    def _group_node_constraints(self, candidate_statements):
        result = []
        already_visited = set()
        for i in range(0, len(candidate_statements)):
            a_statement = candidate_statements[i]
            if a_statement.st_property == self._instantiation_property_str or self._is_a_literal(a_statement.st_type):
                result.append(a_statement)
                already_visited.add(a_statement)
            else:  # a_statement.st_property != self._instantiation_property_str and it is not a literal obj
                if a_statement not in already_visited:
                    already_visited.add(a_statement)
                    group_to_decide = MergeableConstraints(initial_constraint=a_statement,
                                                           statement_serializer_factory=self._statement_serializer_factory,
                                                           namespaces_dict=self._namespaces_dict)

                    result.append(self._find_and_merge_potentially_swapped_constraints(
                        mergeable_constraints=group_to_decide,
                        already_visited=already_visited,
                        all_original_statements=candidate_statements,
                        target_index_original_statements=i+1,
                    ))
        return result

    def _find_and_merge_potentially_swapped_constraints(self, mergeable_constraints,
                                                        already_visited,
                                                        all_original_statements,
                                                        target_index_original_statements):
        self._find_all_candidates_to_merge_swapped_constraints_at_node_level(
            mergeable_constraints=mergeable_constraints,
            already_visited=already_visited,
            all_original_statements=all_original_statements,
            target_index_original_statements=target_index_original_statements
        )
        return self._merge_swapped_constraints_at_node_level(  # TODO
            group_to_merge=mergeable_constraints
        )

    def _find_all_candidates_to_merge_swapped_constraints_at_node_level(self,
                                                                        mergeable_constraints,
                                                                        already_visited,
                                                                        all_original_statements,
                                                                        target_index_original_statements):
        for j in range(target_index_original_statements, len(all_original_statements)):
            if self._statements_have_same_prop_and_are_node_type(mergeable_constraints.get(0),  # todo check have_same_prop_behaviour
                                                                 all_original_statements[j]):
                mergeable_constraints.add_constraint(all_original_statements[j])
                already_visited.add(all_original_statements[j])
        # No need to return anything, modifying the received parameters.

    def _statements_have_same_prop_and_are_node_type(self, original_sentence,
                                                           target_sentence):
        # In this context, we can assume that the original one does not point to a literal
        if target_sentence.st_type in [IRI_ELEM_TYPE, BNODE_ELEM_TYPE] or target_sentence.st_type.startswith(STARTING_CHAR_FOR_SHAPE_NAME):
            return original_sentence.st_property == target_sentence.st_property
        return False

    def _merge_swapped_constraints_at_node_level(self, group_to_merge):
        if len(group_to_merge) == 1:
            return group_to_merge.get(0)
        group_to_merge.sort()
        return group_to_merge.merge_group(disable_or=self._disable_or_statements,
                                          redundant_or_allowed=self._allow_redundant_or)



    def _statements_without_shapes_to_remove(self, original_statements, shape_names_to_remove):
        new_statements = []
        for a_statement in original_statements:
            if not a_statement.st_type in shape_names_to_remove:
                new_statements.append(a_statement)
        return new_statements

    def _is_a_literal(self, node_kind_str):
        if node_kind_str.startswith(STARTING_CHAR_FOR_SHAPE_NAME):
            return False
        if node_kind_str in [IRI_ELEM_TYPE, BNODE_ELEM_TYPE]:
            return False
        return True


    def _shape_name(self, class_key):
        if class_key.startswith("<"):
            return STARTING_CHAR_FOR_SHAPE_NAME + class_key
        return f"{STARTING_CHAR_FOR_SHAPE_NAME}<{self._shape_names_dict[class_key]}>"

class MergeableConstraints(object):
    """
    Internal class used to handle constraints created during the voting process that should be merged into a single one.
    """
    def __init__(self, initial_constraint=None, statement_serializer_factory=None, namespaces_dict=None):
        self._constraints = []
        self._bnode_constraint = None
        self._shape_constraints = None
        self._iri_constraint = None
        self._dominant_constraint = None
        self._disable_or = True
        self._redundant_or_enabled = False
        self._statement_serializer_factory = statement_serializer_factory
        self._namespaces_dict = namespaces_dict
        if initial_constraint is not None:
            self.add_constraint(initial_constraint)

    def add_constraint(self, statement):
        self._constraints.append(statement)
        if statement.st_type == BNODE_ELEM_TYPE:
            self._bnode_constraint = statement
        elif statement.st_type == IRI_ELEM_TYPE:
            self._iri_constraint = statement
        else:
            if self._shape_constraints is None:
                self._shape_constraints = []
            self._shape_constraints.append(statement)


    @property
    def has_bnodes(self):
        return self._bnode_constraint is not None

    @property
    def has_iri_constraint(self):
        return self._iri_constraint is not None

    @property
    def has_shape_constraints(self):
        return self._shape_constraints is not None

    def get(self, index):
        return self._constraints[index]

    def __len__(self):
        return len(self._constraints)

    def sort(self):
        self._constraints.sort(reverse=True, key=lambda x: x.probability)
        if self._shape_constraints is not None:
            self._shape_constraints.sort(reverse=True, key=lambda x: x.probability)

    def constraints(self):
        for a_constaint in self._constraints:
            yield a_constaint

    def merge_group(self, disable_or, redundant_or_allowed):
        self.sort()
        self._disable_or = disable_or
        self._redundant_or_enabled = redundant_or_allowed
        if self.has_bnodes:
            self._bnode_merging_strategy()
        else:
            self._no_bnode_merging_strategy()
        self._merge_content_in_single_statement()
        return self._dominant_constraint

    def _bnode_merging_strategy(self):
        if self.has_iri_constraint:  # iri + bnod = shape
            if len(self._shape_constraints) == 1 \
                    and self._iri_constraint.n_occurences + self._bnode_constraint.n_occurences \
                    == self._shape_constraints[0].n_occurences:
                self._promote_to_dominant(self._shape_constraints[0])
            else: # there are iri, bnode, and shape, not composed
                self._add_dominant(Statement(st_property=self._bnode_constraint.st_property,
                                             st_type=NONLITERAL_ELEM_TYPE,
                                             n_occurences=self._bnode_constraint.n_occurences + self._iri_constraint.n_occurences,
                                             is_inverse=self._bnode_constraint.is_inverse,
                                             probability=self._bnode_constraint.probability + self._iri_constraint.probability,
                                             cardinality=self._most_general_cardinality(self._bnode_constraint.cardinality,
                                                                                        self._iri_constraint.cardinality),
                                             serializer_object=self._statement_serializer_factory.get_base_serializer(is_inverse=self._bnode_constraint.is_inverse)
                                             ))
        elif len(self._shape_constraints) != 0 \
                and self._shape_constraints[0].n_occurences == self._bnode_constraint.n_occurences:
                # Case of at least a shape being used exactyl as many times as BNODE
            self._promote_to_dominant(self._shape_constraints[0])
        else:  # No IRI and no shape is used enough, BNODE should subsume everything.
            self._promote_to_dominant(self._bnode_constraint)



    def _no_bnode_merging_strategy(self):
        if len(self._shape_constraints) == 0 or self._shape_constraints[0].n_occurences < self._iri_constraint.n_occurences:
            self._promote_to_dominant(self._iri_constraint)
        else:
            self._promote_to_dominant(self._shape_constraints[0])

    def _merge_content_in_single_statement(self):
        self._tune_dominant_constraint_wrt_or_config()
        self._feed_dominant_constraint_with_comments()

    def _feed_dominant_constraint_with_comments(self):
        if self._bnode_constraint is not None:
            self._dominant_constraint.add_comment(
                AbstractShexingStrategy._turn_statement_into_comment(self._bnode_constraint,
                                                                     self._namespaces_dict)
            )
            if self._iri_constraint is not None:  # Add IRI one only if there are both bnodes and iris.
                self._dominant_constraint.add_comment(
                    AbstractShexingStrategy._turn_statement_into_comment(self._iri_constraint,
                                                                         self._namespaces_dict)
                )
        for a_constraint in self._shape_constraints:
            if self._dominant_constraint != a_constraint:
                self._dominant_constraint.add_comment(
                    AbstractShexingStrategy._turn_statement_into_comment(a_constraint,
                                                                         self._namespaces_dict)
                )

    def _tune_dominant_constraint_wrt_or_config(self):
        if self._disable_or:
            if self._dominant_constraint.st_type not in [IRI_ELEM_TYPE, BNODE_ELEM_TYPE, NONLITERAL_ELEM_TYPE] \
                    and len(self._shape_constraints) > 0 \
                    and self._shape_constraints[0].n_occurences == self._dominant_constraint:
                if self._iri_constraint is not None:
                    self._promote_to_dominant(self._iri_constraint)
                else:  # It must be a BNODE
                    self._promote_to_dominant(self._bnode_constraint)
        else:  # or allowed
            st_types = []
            if self._redundant_or_enabled:
                if self._dominant_constraint not in self._shape_constraints:
                    st_types.append(self._dominant_constraint.st_type)
                st_types = st_types + [a_constraint.st_type for a_constraint in self._shape_constraints]
            elif not self._redundant_or_enabled and self._dominant_constraint in self._shape_constraints:
                st_types = st_types + [a_constraint.st_type for a_constraint in self._shape_constraints]
            if len(st_types) > 1:
                self._dominant_constraint = FixedPropChoiceStatement(
                    st_property=self._dominant_constraint.st_property,
                    st_types=st_types,
                    cardinality=self._dominant_constraint.cardinality,
                    probability=self._dominant_constraint.probability,
                    n_occurences=self._dominant_constraint.n_occurences,
                    serializer_object=self._statement_serializer_factory.get_choice_serializer(
                        is_inverse=self._dominant_constraint.is_inverse
                    ),
                    is_inverse=self._dominant_constraint.is_inverse)





    def _most_general_cardinality(self, a_card1, a_card2):
        if POSITIVE_CLOSURE in (a_card1, a_card2) or a_card1 != a_card2:
            return POSITIVE_CLOSURE
        else:
            return a_card1
    def _add_dominant(self, statement):
        self._dominant_constraint = statement

    def _promote_to_dominant(self, statement):
        self._dominant_constraint = statement
        self._constraints.remove(statement)

    def _demote_dominant_to_plain(self):
        if self._dominant_constraint.st_type not in [IRI_ELEM_TYPE, BNODE_ELEM_TYPE, NONLITERAL_ELEM_TYPE]:
            self.add_constraint(self._dominant_constraint)
            self.sort()
        self._dominant_constraint = None




import unittest
from shexer.shaper import Shaper
from test.const import G1_NT, BASE_FILES, default_namespaces, G1_ALL_CLASSES_NO_COMMENTS
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth

from shexer.consts import TURTLE



_BASE_DIR = BASE_FILES + "no_slash_prefixed" + pth.sep  # We just need something with another instantiation property


class TestNoSharpNorSlashDueToPrefixing(unittest.TestCase):

    def test_shorter_prefixes_g1(self):
        shaper = Shaper(
            graph_file_input=G1_NT,
            namespaces_dict={"http://example.org/": "ex",
            "http://www.w3.org/XML/1998/": "xml",
            "http://www.w3.org/1999/02/": "rdf",
            "http://www.w3.org/2000/01#": "rdfs",
            "http://www.w3.org/2001#": "xsd",
            "http://xmlns.com/foaf/": "foaf"
            },
            all_classes_mode=True,
            disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g1_no_prefix_except_shapes.shex",
                                                      str_target=str_result))


import re

from shexer.core.profiling.class_profiler import RDF_TYPE_STR

from shexer.model.property import Property
from shexer.utils.uri import remove_corners, prefixize_uri_if_possible
from shexer.utils.shapes import prefixize_shape_name_if_possible
from shexer.io.shex.formater.consts import SPACES_LEVEL_INDENTATION
from shexer.io.wikidata import wikidata_annotation
from shexer.io.file import read_file
from shexer.consts import RATIO_INSTANCES, ABSOLUTE_INSTANCES, MIXED_INSTANCES, ALL_EXAMPLES, \
    SHAPE_EXAMPLES, CONSTRAINT_EXAMPLES, EXAMPLE_CONFORMANCE_PROP, ABSOLUTE_COUNT_PROP
from shexer.io.shex.formater.consts import SPACES_GAP_BETWEEN_TOKENS

from wlighter import SHEXC_FORMAT

_MODES_REPORT_INSTANCES = [ABSOLUTE_INSTANCES, MIXED_INSTANCES]
_EXAMPLE_CONSTRAINT_TEMPLATE = '// rdfs:comment {} ;'
_EXAMPLE_INSTANCE_TEMPLATE = " // {} {}"

_INIT_URI_PATTERN = re.compile("http[s]?\://")


class ShexSerializer(object):

    def __init__(self, target_file, shapes_list, namespaces_dict=None, string_return=False,
                 instantiation_property_str=RDF_TYPE_STR, disable_comments=False, wikidata_annotation=False,
                 instances_report_mode=RATIO_INSTANCES, detect_minimal_iri=False, shape_example_features=None,
                 examples_mode=None, inverse_paths=False, example_constraint_prop=EXAMPLE_CONFORMANCE_PROP,
                 comments_to_annotations=False, absolute_counts_prop=ABSOLUTE_COUNT_PROP):
        self._target_file = target_file
        self._shapes_list = shapes_list
        self._lines_buffer = []
        self._namespaces_dict = namespaces_dict if namespaces_dict is not None else {}
        self._string_return = string_return
        self._instantiation_property_str = self._decide_instantiation_property(instantiation_property_str)
        self._disable_comments = disable_comments
        self._wikidata_annotation = wikidata_annotation
        self._instances_report_mode = instances_report_mode
        self._detect_minimal_iri = detect_minimal_iri
        self._examples_mode = examples_mode
        self._shape_example_features = shape_example_features
        self._inverse_paths = inverse_paths
        self._example_constraint_prop = example_constraint_prop
        self._comments_to_annotations = comments_to_annotations
        self._absolute_counts_property = absolute_counts_prop

        self._string_result = ""

    def serialize_shapes(self):

        self._reset_target_file()
        self._serialize_namespaces()
        for a_shape in self._shapes_list:
            self._serialize_shape(a_shape)
        self._flush()
        if self._wikidata_annotation:
            self._annotate_wikidata_ids_in_result()
        if self._string_return:
            return self._string_result

    @staticmethod
    def _decide_instantiation_property(instantiation_property_str):
        if instantiation_property_str == None:
            return RDF_TYPE_STR
        if type(instantiation_property_str) == Property:
            return str(instantiation_property_str)
        if type(instantiation_property_str) == str:
            return remove_corners(a_uri=instantiation_property_str,
                                  raise_error_if_no_corners=False)
        raise ValueError("Unrecognized param type to define instantiation property")

    def _annotate_wikidata_ids_in_result(self):
        self._string_result = wikidata_annotation(raw_input=self._get_raw_input_for_wikidata_annotation(),
                                                  string_return=self._string_return,
                                                  out_file=self._target_file,
                                                  format=SHEXC_FORMAT,
                                                  rdfs_comments=True)

    def _get_raw_input_for_wikidata_annotation(self):
        if self._string_return:
            return self._string_result
        return read_file(self._target_file)

    def _serialize_namespaces(self):
        for a_namespace in self._namespaces_dict:
            self._write_line(self._prefix_line(a_namespace), 0)
        self._write_line("", 0)

    def _prefix_line(self, namespace_key):
        return "PREFIX " + self._namespaces_dict[namespace_key] + ": <" + namespace_key + ">"

    def _serialize_empty_namespace(self):
        self._write_line("PREFIX : <http://weso.es/shapes/>")

    def _serialize_shape(self, a_shape):
        self._serialize_shape_name(a_shape)
        self._serialize_opening_of_rules()
        self._serialize_shape_rules(a_shape)
        self._serialize_closure_of_rules(a_shape)
        self._serialize_shape_gap()

    def _flush(self):
        self._write_lines_buffer()

    def _write_line(self, a_line, indent_level=0):
        self._lines_buffer.append(self._indentation_spaces(indent_level) + a_line + "\n")
        if len(self._lines_buffer) >= 5000:
            self._write_lines_buffer()
            self._lines_buffer = []

    def _reset_target_file(self):
        if self._string_return:
            return
        with open(self._target_file, "w") as out_stream:
            out_stream.write("")  # Is this necessary? maybe enough to open it in 'w' mode?

    def _write_lines_buffer(self):
        if self._string_return:
            self._string_result += "".join(self._lines_buffer)
        else:
            with open(self._target_file, "a") as out_stream:
                for a_line in self._lines_buffer:
                    out_stream.write(a_line)

    def _indentation_spaces(self, indent_level):
        result = ""
        for i in range(0, indent_level):
            result += SPACES_LEVEL_INDENTATION
        return result

    def _serialize_shape_rules(self, a_shape):
        if a_shape.n_statements == 0:
            return
        self._tune_statement_examples_if_needed(a_shape)
        statements = [a_statement for a_statement in a_shape.yield_statements()]

        for i in range(0, len(statements) - 1):
            for line_indent_tuple in statements[i]. \
                    get_tuples_to_serialize_line_indent_level(is_last_statement_of_shape=False,
                                                              namespaces_dict=self._namespaces_dict):
                self._write_line(a_line=line_indent_tuple[0],
                                 indent_level=line_indent_tuple[1])
        for line_indent_tuple in statements[len(statements) - 1]. \
                get_tuples_to_serialize_line_indent_level(is_last_statement_of_shape=True,
                                                          namespaces_dict=self._namespaces_dict):
            self._write_line(a_line=line_indent_tuple[0],
                             indent_level=line_indent_tuple[1])

    def _tune_statement_examples_if_needed(self, a_shape):
        if self._examples_mode not in [ALL_EXAMPLES, CONSTRAINT_EXAMPLES]:
            return
        self._add_statement_examples(a_shape)

    def _add_statement_examples(self, a_shape):
        if self._comments_to_annotations:
            for a_statement in a_shape.yield_statements():
                if a_statement.st_property != self._instantiation_property_str:
                    a_statement.add_annotation(predicate=self._example_constraint_prop,
                                               obj=self._get_node_constraint_example_no_inverse(
                                                   a_shape,
                                                   a_statement
                                               ) if not self._inverse_paths
                                               else
                                               self._get_node_constraint_example_inverse(a_shape, a_statement),
                                               insert_first=True
                                               )
        else:
            for a_statement in a_shape.yield_statements():
                if a_statement.st_property != self._instantiation_property_str:
                    comment = _EXAMPLE_CONSTRAINT_TEMPLATE.format(
                        self._turn_str_comment_into_proper_rdf(
                            self._get_node_constraint_example_no_inverse(a_shape, a_statement) if not self._inverse_paths
                            else self._get_node_constraint_example_inverse(a_shape, a_statement)
                        )
                    )

                    a_statement.add_comment(comment, insert_first=True)

    def _turn_str_comment_into_proper_rdf(self, str_object_to_transform):
        """
        If it's a whole prefixed URI, return it as it is.
        If it is a literal, surround it with quotes.
        If it starts with http(s)://, surround it with corners.
        """

        if " " not in str_object_to_transform and "".count(":") == 1:
            return str_object_to_transform
        elif _INIT_URI_PATTERN.match(str_object_to_transform):
            prefixed = prefixize_uri_if_possible(str_object_to_transform, namespaces_prefix_dict=self._namespaces_dict,
                                                 corners=False)
            if prefixed == str_object_to_transform:
                return "<" + str_object_to_transform + ">"
            else:
                return prefixed
        else:
            return '"' + str_object_to_transform + '"'

    def _get_node_constraint_example_no_inverse(self, shape, statement):
        candidate = self._shape_example_features.get_constraint_example(shape_id=shape.class_uri,
                                                                        prop=statement.st_property)
        if candidate.startswith("http"):  # Let's assume this means that it is an URI
            candidate = prefixize_uri_if_possible(target_uri=candidate,
                                                  namespaces_prefix_dict=self._namespaces_dict,
                                                  corners=False)
        return candidate

    def _get_node_constraint_example_inverse(self, shape, statement):
        candidate = self._shape_example_features.get_constraint_example(shape_id=shape.class_uri,
                                                                        prop=statement.st_property,
                                                                        inverse=statement.is_inverse)
        if candidate.startswith("https://"):  # Let's assume this means that it is a URI
            candidate = prefixize_uri_if_possible(target_uri=candidate,
                                                  namespaces_prefix_dict=self._namespaces_dict,
                                                  corners=False)
        return candidate

    def _serialize_shape_name(self, a_shape):
        self._write_line(

            prefixize_shape_name_if_possible(a_shape_name=a_shape.name,
                                             namespaces_prefix_dict=self._namespaces_dict) +
            self._minimal_iri(a_shape=a_shape) +
            self._instance_count(a_shape)
        )

    def _serialize_example(self, a_shape):
        if self._examples_mode not in [ALL_EXAMPLES, SHAPE_EXAMPLES]:
            return ""
        candidate = self._shape_example_features.shape_example(shape_id=a_shape.class_uri)
        prefixed_example = prefixize_uri_if_possible(candidate, namespaces_prefix_dict=self._namespaces_dict, corners=False)
        prefixed_property = prefixize_uri_if_possible(self._example_constraint_prop, namespaces_prefix_dict=self._namespaces_dict, corners=False)
        return _EXAMPLE_INSTANCE_TEMPLATE.format(
            prefixed_property if prefixed_property != self._example_constraint_prop else f'<{self._example_constraint_prop}>',
            prefixed_example if prefixed_example != candidate else f'<{candidate}>'
        )

    def _serialize_shape_count(self, a_shape):
        if self._instances_report_mode not in [ABSOLUTE_INSTANCES, MIXED_INSTANCES] or self._disable_comments:
            return ""
        prefixed_property = prefixize_uri_if_possible(self._absolute_counts_property,
                                                      namespaces_prefix_dict=self._namespaces_dict,
                                                      corners=False)
        return _EXAMPLE_INSTANCE_TEMPLATE.format(
            prefixed_property if prefixed_property != self._absolute_counts_property else f'<{self._absolute_counts_property}>',
            str(a_shape.n_instances)
        )

    def _minimal_iri(self, a_shape):
        if not self._detect_minimal_iri or self._shape_example_features.shape_min_iri(a_shape.class_uri) is None:
            return ""
        return "  [<{}>~]  AND".format(self._shape_example_features.shape_min_iri(a_shape.class_uri))

    def _instance_count(self, a_shape):
        result = "   # {} instance{}.".format(a_shape.n_instances,
                                            "" if a_shape.n_instances == 1 else "s") \
            if self._instances_report_mode in _MODES_REPORT_INSTANCES and not self._disable_comments and not self._comments_to_annotations \
            else ""
        return result

    def _serialize_opening_of_rules(self):
        self._write_line("{")

    def _serialize_closure_of_rules(self, a_shape):
        self._write_line("}" + self._serialize_example(a_shape=a_shape) + SPACES_GAP_BETWEEN_TOKENS +  self._serialize_shape_count(a_shape))

    def _serialize_shape_gap(self):
        self._write_line("")
        self._write_line("")



from shexer.io.shex.formater.statement_serializers.frequency_strategy.base_frequency_strategy import BaseFrequencyStrategy
from shexer.io.shex.formater.statement_serializers.frequency_strategy.abs_freq_serializer import AbsFreqSerializer
from shexer.io.shex.formater.statement_serializers.frequency_strategy.ratio_freq_serializer import RatioFreqSerializer

class MixedFrequencyStrategy(BaseFrequencyStrategy):

    def __init__(self, frequency_ratio_property, frequency_absolute_property, namespaces_dict, decimals=-1):
        super().__init__(frequency_ratio_property, frequency_absolute_property, namespaces_dict)
        self._abs_strategy = AbsFreqSerializer(frequency_ratio_property, frequency_absolute_property, namespaces_dict)
        self._ratio_strategy = RatioFreqSerializer(frequency_ratio_property, frequency_absolute_property, namespaces_dict, decimals)

    def serialize_frequency(self, statement):
        # The abs_strategy return a trailing dot that we want to skip. That why I use slicing here
        return self._ratio_strategy.serialize_frequency(statement) + \
               " (" + self._abs_strategy.serialize_frequency(statement)[:-1] + ")."

    def annotations_for_frequency(self, statement):
        return self._ratio_strategy.annotations_for_frequency(statement) + \
            self._abs_strategy.annotations_for_frequency(statement)





from shexer.model.const_elem_types import *

_VALID_MACROS = [IRI_ELEM_TYPE, LITERAL_ELEM_TYPE, DOT_ELEM_TYPE, BNODE_ELEM_TYPE]

class Macro(object):
    def __init__(self, macro_const):
        if macro_const not in _VALID_MACROS:
            # print(_VALID_MACROS)
            # print(macro_const == DOT_ELEM_TYPE)
            raise ValueError("Not recognized macro: " + macro_const)
        self._macro_representation = macro_const

    def __str__(self):
        return self._macro_representation

    @property
    def elem_type(self):
        return self._macro_representation

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        return str(other) == str(self)

    def __ne__(self, other):
        return not self.__neq__(other)


from xz import open as xzopen


class XzFileLineReader(object):

    def __init__(self, xz_file):
        self._xz_file = xz_file

    def read_lines(self):
        with xzopen(self._xz_file, "r") as in_stream:
            for a_line in in_stream:
                yield a_line.decode("utf-8")


import unittest
from shexer.shaper import Shaper
from test.const import G1, BASE_FILES, G1_NT, default_namespaces, G1_ALL_CLASSES_NO_COMMENTS
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth
from shexer.consts import TURTLE

_BASE_DIR = BASE_FILES + "general" + pth.sep

class TestGraphFileInput(unittest.TestCase):

    def test_some_format(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))


    def test_no_format(self):  # Should be nt
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1_NT,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))


import unittest
from shexer.shaper import Shaper
from shexer.consts import NT
from test.const import G1_NT, BASE_FILES, BASE_FILES_GENERAL, default_namespaces
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth

_BASE_DIR = BASE_FILES + "instances_file_input" + pth.sep

class TestInstancesFileInput(unittest.TestCase):

    def test_all_classes_all_instances(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1_NT,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=NT,
                        disable_comments=True,
                        instances_file_input=_BASE_DIR + "g1_all_instances.nt")
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=BASE_FILES_GENERAL + "g1_all_classes_no_comments.shex",
                                                      str_target=str_result))

    def test_all_classes_some_instances(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1_NT,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=NT,
                        disable_comments=True,
                        instances_file_input=_BASE_DIR + "g1_some_instances.nt")
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "all_classes_some_instances.shex",
                                                      str_target=str_result))

    def test_some_classes_some_instances(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person"],
                        graph_file_input=G1_NT,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=NT,
                        disable_comments=True,
                        instances_file_input=_BASE_DIR + "g1_some_instances.nt")
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "some_classes_some_instances.shex",
                                                      str_target=str_result))

    def test_some_classes_all_instances(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person"],
                        graph_file_input=G1_NT,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=NT,
                        disable_comments=True,
                        instances_file_input=_BASE_DIR + "g1_all_instances.nt")
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "some_classes_all_instances.shex",
                                                      str_target=str_result))


import unittest
from shexer.shaper import Shaper
from test.const import G1, BASE_FILES
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth

_BASE_DIR = BASE_FILES + "target_classes" + pth.sep

class TestTargetClasses(unittest.TestCase):

    def test_one_target(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person"],
                        graph_file_input=G1,
                        all_classes_mode=False,
                        input_format="turtle",
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "one_target.shex",
                                                      str_target=str_result,
                                                      strict=False))


    def test_several_targets(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1,
                        all_classes_mode=False,
                        input_format="turtle",
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "two_targets.shex",
                                                      str_target=str_result,
                                                      strict=False))


    def test_one_target_prefixed_targets(self):
        shaper = Shaper(target_classes=["foaf:Person"],
                        namespaces_dict={"http://xmlns.com/foaf/0.1/" : "foaf"},
                        graph_file_input=G1,
                        all_classes_mode=False,
                        input_format="turtle",
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "one_target.shex",
                                                      str_target=str_result,
                                                      strict=False))







import argparse

from shexer.consts import DEFAULT_NAMESPACES, SHACL_TURTLE, NT
from shexer.shaper import Shaper


def define_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input_kg_path", required=True, help="Input path to the KG")
    parser.add_argument("-o", "--output_shapes_path", required=True, help="Output path for the generated shapes in Turtle")
    parser.add_argument("-f", "--format", required=False,
                        help="KG input format. Values: nt/tsv_spo/turtle/turtle_iter/xml/n3/json-ld (NT default)")
    return parser


def main():
    args = define_args().parse_args()
    if args.format:
        kg_format = args.format
    else:
        kg_format = NT

    shaper = Shaper(
        graph_file_input=args.input_kg_path,
        namespaces_dict=DEFAULT_NAMESPACES,
        all_classes_mode=True,
        input_format=kg_format)
    shaper.shex_graph(output_file=args.output_shapes_path, output_format=SHACL_TURTLE)


from shexer.core.profiling.class_profiler import ClassProfiler

from shexer.consts import SHAPES_DEFAULT_NAMESPACE
from shexer.core.profiling.consts import RDF_TYPE_STR
from shexer.utils.log import log_msg
from shexer.model.graph.endpoint_sgraph import EndpointSGraph
from shexer.utils.triple_yielders import tune_token, tune_prop
from shexer.core.instances.pconsts import _S, _P, _O

class FederatedSourceClassProfiler(ClassProfiler):

    def __init__(self, triples_yielder, instances_dict, instantiation_property_str=RDF_TYPE_STR,
                 remove_empty_shapes=True, original_target_classes=None, original_shape_map=None,
                 shapes_namespace=SHAPES_DEFAULT_NAMESPACE, inverse_paths=False, detect_minimal_iri=False,
                 examples_mode=None, list_of_federated_objects=None, namespaces_dict=None):
        super().__init__(triples_yielder, instances_dict, instantiation_property_str, remove_empty_shapes,
                         original_target_classes, original_shape_map, shapes_namespace, inverse_paths,
                         detect_minimal_iri, examples_mode, namespaces_dict)

        self._list_of_federated_objects = list_of_federated_objects


    def _yield_relevant_triples(self):
        for a_triple in self._triples_yielder.yield_triples():
            if self._strategy.is_a_relevant_triple(a_triple):
                yield a_triple

        for a_fed_source in self._list_of_federated_objects:
            for a_triple in self._yield_triples_of_fed_source(a_fed_source):
                yield a_triple

    def _yield_triples_of_fed_source(self, fed_source):
        end_graph = EndpointSGraph(fed_source.endpoint_url, store_locally=False)
        if not self._inverse_paths:
            triples = self._direct_target_triples_of_fed_source(end_graph, fed_source)
        else:
            triples = self._direct_and_inverse_target_triples_of_fed_source(end_graph, fed_source)
        for a_triple in triples:
            yield (
                tune_token(a_triple[_S]),
                tune_prop(a_triple[_P]),
                tune_token(a_triple[_O])
            )

    def _direct_target_triples_of_fed_source(self, end_graph, fed_source):
        triples = set()
        for a_target_node in fed_source.set_of_instances:
            for a_triple in end_graph.yield_p_o_triples_of_an_s(a_target_node):
                triples.add(a_triple)
        return triples

    def _direct_and_inverse_target_triples_of_fed_source(self, end_graph, fed_source):
        triples = set()
        for a_target_node in fed_source.set_of_instances:
            for a_triple in end_graph.yield_p_o_triples_of_an_s(a_target_node):
                triples.add(a_triple)
            for a_triple in end_graph.yield_s_p_triples_of_an_o(a_target_node):
                triples.add(a_triple)
        return triples



from shexer.utils.log import log_msg
from shexer.utils.literal import there_is_arroba_after_last_quotes
from shexer.utils.triple_yielders import tune_prop, tune_token  # , check_if_property_belongs_to_namespace_list
from shexer.io.graph.yielder.base_triples_yielder import BaseTriplesYielder
from shexer.utils.exception import ParseError



class NtTriplesYielder(BaseTriplesYielder):

    def __init__(self, source_file=None, allow_untyped_numbers=False, raw_graph=None,
                 compression_mode=None, zip_base_archive=None, silent_error=False):

        super(NtTriplesYielder, self).__init__()
        self._source_file = source_file
        self._raw_graph = raw_graph
        self._triples_count = 0
        self._error_triples = 0
        self._allow_untyped_numbers = allow_untyped_numbers
        self._silent_error = silent_error
        self._line_reader = self._decide_line_reader(source_file=source_file,
                                                     raw_graph=raw_graph,
                                                     compression_mode=compression_mode,
                                                     zip_base_archive=zip_base_archive)
        # The following ones are refs to functions. I'm avoiding some comparison here.
        # self.yield_triples = self._yield_triples_not_excluding_namespaces if namespaces_to_ignore is None \
        #     else self._yield_triples_excluding_namespaces

    def yield_triples(self):
        self._reset_count()
        try:
            for a_line in self._line_reader.read_lines():
                a_line = a_line.strip()
                if a_line != "":
                    tokens = self._look_for_tokens(a_line)
                    if len(tokens) != 3:
                        if self._silent_error:
                            self._error_triples += 1
                            log_msg(verbose=False, msg="This line was discarded: " + a_line)
                        else:
                            raise ParseError(f"Line: '{a_line}'")
                    else:
                        yield (tune_token(a_token=tokens[0]),
                               tune_prop(a_token=tokens[1]),
                               tune_token(a_token=tokens[2],
                                          allow_untyped_numbers=self._allow_untyped_numbers))
                        self._triples_count += 1
        except BaseException as e:
            raise ParseError(f"Error while parsing. {e}") from e

    def _look_for_tokens(self, str_line):
        result = []
        current_first_index = 0
        while current_first_index != len(str_line):
            if str_line[current_first_index] == "<":
                last_index = self._look_for_last_index_of_uri_token(str_line, current_first_index)
                result.append(str_line[current_first_index:last_index + 1])
                current_first_index = last_index + 1
            elif str_line[current_first_index] == '"':
                last_index = self._look_for_last_index_of_literal_token(str_line, current_first_index)
                result.append(str_line[current_first_index:last_index + 1])
                current_first_index = last_index + 1
            elif str_line[current_first_index] == '_':
                last_index = self._look_for_last_index_of_bnode_token(str_line, current_first_index)
                result.append(str_line[current_first_index:last_index + 1])
                current_first_index = last_index + 1
            elif str_line[current_first_index] == '.':
                break

            elif str_line[current_first_index].isnumeric():
                last_index = self._look_for_last_index_of_unlabelled_number_token(str_line, current_first_index)
                result.append(str_line[current_first_index:last_index + 1])
                current_first_index = last_index + 1
            else:
                current_first_index += 1

        return result

    def _look_for_last_index_of_uri_token(self, target_str, first_index):
        target_substring = target_str[first_index:]
        index_sub = target_substring.find(">")
        return index_sub + (len(target_str) - len(target_substring))

    def _look_for_last_index_of_bnode_token(self, target_str, first_index):
        target_substring = target_str[first_index:]
        index_sub = target_substring.find(" ")
        return index_sub + (len(target_str) - len(target_substring)) - 1

    def _look_for_last_index_of_unlabelled_number_token(self, target_str, first_index):
        target_substring = target_str[first_index:]
        index_sub = target_substring.find(" ")
        return index_sub + (len(target_str) - len(target_substring)) - 1

    def _look_for_last_index_of_literal_token(self, target_str, first_index):
        target_substring = target_str[first_index:]

        if there_is_arroba_after_last_quotes(target_substring):  # String labelled with language
            return target_substring[target_substring.rfind("@"):].find(" ") - 1 + target_str.rfind("@")
        elif "^^" not in target_substring:  # Not typed
            success = False
            index_of_quotes = 1
            while not success:
                index_of_second_quotes = target_substring[index_of_quotes + 1:].find('"') + index_of_quotes + 1
                if target_substring[index_of_second_quotes - 1] != "\\":
                    success = True
                elif target_substring[index_of_second_quotes - 2] == "\\":  # Case of escaped slash "\\"
                    success = True
                index_of_quotes = index_of_second_quotes
            return index_of_quotes + (len(target_str) - len(target_substring))
        else:  # Typed
            return target_substring[target_substring.find("^^"):].find(" ") - 1 + target_str.find("^^")

    @property
    def yielded_triples(self):
        return self._triples_count

    @property
    def error_triples(self):
        return self._error_triples

    def _reset_count(self):
        self._error_triples = 0
        self._triples_count = 0




SPACES_GAP_FOR_FREQUENCY = "          "
SPACES_GAP_BETWEEN_TOKENS = "  "
TARGET_LINE_LENGHT = 60
SPACES_LEVEL_INDENTATION = "   "
COMMENT_INI = "# "
POSITIVE_CLOSURE = "+"
KLEENE_CLOSURE = "*"
OPT_CARDINALITY = "?"
SHAPE_LINK_CHAR = "@"
ANNOTATION_BEGIN = "//"



XSD_NAMESPACE = "http://www.w3.org/2001/XMLSchema#"
XSD_PREFIX = "xsd"

RDF_SYNTAX_NAMESPACE = "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
RDF_PREFIX = "rdf"
RDF_TYPE = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"

DT_NAMESPACE = "http://dbpedia.org/datatype/"
DT_PREFIX = "dt"

OPENGIS_NAMESPACE = "http://www.opengis.net/ont/geosparql#"
OPENGIS_PREFIX = "geo"

LANG_STRING_TYPE = "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"
STRING_TYPE = "http://www.w3.org/2001/XMLSchema#string"
FLOAT_TYPE = "http://www.w3.org/2001/XMLSchema#float"
INTEGER_TYPE = "http://www.w3.org/2001/XMLSchema#integer"

from shexer.model.shape import STARTING_CHAR_FOR_SHAPE_NAME

def _count_prior_backslashes(an_str, quote_pos):
    """
    We assume that there is at least a backslash at an_str[pos-1], so pos-1 is a non-negative index of an_str
    """
    counter = 1
    quote_pos -= 2
    while quote_pos >= 0:
        if an_str[quote_pos] == "\\":
            counter += 1
        else:
            return counter
        quote_pos -= 1
    return counter

def find_next_unescaped_quotes(target_str, start_index):
    pos = target_str.find('"', start_index)
    while pos != -1:
        if target_str[pos - 1] != "\\":
            return pos  # not escaped
        # if pos >= 2 and target_str[pos-2] == '\\':
        #     return pos  # the scape is scaped, so not escaped
        if _count_prior_backslashes(an_str=target_str,
                                    quote_pos=pos) % 2 == 0:
            return pos  # the scape is scaped, so not escaped
        pos = target_str.find('"', pos + 1)
    if pos == -1:
        raise ValueError("Is this line malformed? Can`t find quotes matching: " + target_str)

def there_is_arroba_after_last_quotes(target_str):
    if target_str.rfind(STARTING_CHAR_FOR_SHAPE_NAME) > target_str.rfind('"'):
        return True
    return False

def decide_literal_type(a_literal, base_namespace=None):
    if there_is_arroba_after_last_quotes(a_literal):
        return LANG_STRING_TYPE
    elif "\"^^" not in a_literal:
        return STRING_TYPE
    elif "xsd:" in a_literal:
        return XSD_NAMESPACE + a_literal[a_literal.find("xsd:") + 4:]
    elif "rdf:" in a_literal:
        return RDF_SYNTAX_NAMESPACE + a_literal[a_literal.find("rdf:")+ 4:]
    elif "dt:" in a_literal:
        return DT_NAMESPACE + a_literal[a_literal.find("dt:")+ 3:]
    elif "geo:" in a_literal:
        return OPENGIS_NAMESPACE + a_literal[a_literal.find("geo:") + 4:]
    elif XSD_NAMESPACE in a_literal or RDF_SYNTAX_NAMESPACE in a_literal \
            or DT_NAMESPACE in a_literal or OPENGIS_NAMESPACE in a_literal:
        return a_literal[a_literal.find("\"^^")+4:-1]
    elif a_literal.strip().endswith(">"):
        candidate_type = a_literal[a_literal.find("\"^^") + 4:-1]  # plain uri, no corners
        if base_namespace is not None and not candidate_type.startswith("http"):
            return base_namespace + candidate_type
        return candidate_type
    else:
        raise RuntimeError("Unrecognized literal type:" + a_literal)

def parse_literal(an_elem, base_namespace=None):
    closing_quotes = find_next_unescaped_quotes(an_elem, 1)
    content = an_elem[1:closing_quotes].replace("\\\"", "\"")
    elem_type = decide_literal_type(a_literal=an_elem,
                                    base_namespace=base_namespace)
    return content, elem_type

def parse_unquoted_literal(an_elem):
    elem_type = decide_literal_type(an_elem)
    return an_elem, elem_type




from shexer.utils.factories.triple_yielders_factory import produce_shape_map_according_to_input
from shexer.model.graph.rdflib_sgraph import RdflibSgraph


def get_shape_map_if_needed(sm_format, remote_sgraph, namespaces_prefix_dict, target_classes,
                            file_target_classes, shape_map_file, shape_map_raw,
                            instantiation_property, shape_map_already_built=None,
                            rdflib_graph=None, raw_graph=None, source_file_graph=None, input_format=None,
                            limit_remote_instances=-1):
    if shape_map_file is None and shape_map_raw is None:
        return None
    if shape_map_already_built:
        return shape_map_already_built

    sgraph = remote_sgraph if remote_sgraph is not None else RdflibSgraph(rdflib_graph=rdflib_graph,
                                                                          raw_graph=raw_graph,
                                                                          source_file=source_file_graph,
                                                                          format=input_format)

    return produce_shape_map_according_to_input(sm_format=sm_format,
                                                sgraph=sgraph,
                                                namespaces_prefix_dict=namespaces_prefix_dict,
                                                target_classes=target_classes,
                                                file_target_classes=file_target_classes,
                                                shape_map_file=shape_map_file,
                                                shape_map_raw=shape_map_raw,
                                                instantiation_property=instantiation_property,
                                                shape_map_already_built=shape_map_already_built,
                                                limit_remote_instances=limit_remote_instances)



def check_just_one_not_none(*value_refname):
    nones=0
    for a_tuple in value_refname:
        if a_tuple[0] is not None:
            nones += 1
    if nones != 1:
        raise ValueError(error_message_for_non_compatible_references([a_tuple[1] for a_tuple in value_refname]))


def check_one_or_zero_not_none(*value_refname):
    nones=0
    for a_tuple in value_refname:
        if a_tuple[0] is not None:
            nones += 1
    if nones > 1:
        raise ValueError(error_message_for_non_compatible_references(
            list_of_ref_names=[a_tuple[1] for a_tuple in value_refname],
            one_mandatory=False))


def error_message_for_non_compatible_references(list_of_ref_names, one_mandatory=True):
    if one_mandatory:
        return "You must provide one and only one of the following params: " + str(list_of_ref_names)
    return "You must provide one as most of the following params: " + str(list_of_ref_names)






_MIN_IRI_POS = 0
_EXAMPLE_ENTITY_POS = 1
_PROP_FEATURES_POS = 2


_POS_DIRECT = 0
_POS_INVERSE = 1


class ShapeExampleFeaturesDict(object):

    def __init__(self, track_inverse_features):
        self._base_dict = {}
        self._track_inverse_features = track_inverse_features
        self._init_example_tracking_methods()


    def _init_example_tracking_methods(self):
        if self._track_inverse_features:
            self.has_constraint_example = self._has_constraint_example_inverse
            self.set_constraint_example = self._set_constraint_example_inverse
            self.get_constraint_example = self._get_constraint_example_inverse
        else:
            self.has_constraint_example = self._has_constraint_example_no_inverse
            self.set_constraint_example = self._set_constraint_example_no_inverse
            self.get_constraint_example = self._get_constraint_example_no_inverse


    def set_shape_min_iri(self, shape_id, min_iri):
        if shape_id not in self._base_dict:
            self._init_shape(shape_id)
        self._base_dict[shape_id][_MIN_IRI_POS] = min_iri

    def shape_min_iri(self, shape_id):
        return self._base_dict[shape_id][_MIN_IRI_POS]

    def set_shape_example(self, shape_id, example_iri):
        if shape_id not in self._base_dict:
            self._init_shape(shape_id)
        self._base_dict[shape_id][_EXAMPLE_ENTITY_POS] = example_iri

    def shape_example(self, shape_id):
        if shape_id not in self._base_dict:
            return False
        return self._base_dict[shape_id][_EXAMPLE_ENTITY_POS]

    def _init_shape(self, shape_id):
        self._base_dict[shape_id] = [None, None, {} if not self._track_inverse_features else [{}, {}]]


    def has_constraint_example(self, shape_id, prop_id):
        raise NotImplementedError()

    def _has_constraint_example_abstract(self):
        raise NotImplementedError()

    def set_constraint_example(self, shape_id, prop, example):
        raise NotImplementedError()

    def get_constraint_example(self, shape_id, prop):
        raise NotImplementedError()

    def _get_constraint_example_no_inverse(self, shape_id, prop):
        return self._base_dict[shape_id][_PROP_FEATURES_POS][prop]

    def _get_constraint_example_inverse(self, shape_id, prop, inverse):
        return self._base_dict[shape_id][_PROP_FEATURES_POS][_POS_INVERSE if inverse else _POS_DIRECT][prop]

    def _set_constraint_example_no_inverse(self, shape_id, prop_id, example):
        if shape_id not in self._base_dict:
            self._init_shape(shape_id)
        self._base_dict[shape_id][_PROP_FEATURES_POS][prop_id] = self._normalize_example(example)

    def _set_constraint_example_inverse(self, shape_id, prop_id, example, inverse):
        if shape_id not in self._base_dict:
            self._init_shape(shape_id)
        self._base_dict[shape_id][_PROP_FEATURES_POS][_POS_INVERSE if inverse else _POS_DIRECT][prop_id] = self._normalize_example(example)

    def _normalize_example(self, example):
        result = example.replace("\n", "\\n")
        return result.replace("\"", '\\"')

    def _has_constraint_example_no_inverse(self, shape_id, prop_id):
        if shape_id not in self._base_dict:
            return False
        return prop_id in self._base_dict[shape_id][_PROP_FEATURES_POS]

    def _has_constraint_example_inverse(self, shape_id, prop_id, inverse):
        if shape_id not in self._base_dict:
            return False
        return  prop_id in self._base_dict[shape_id][_PROP_FEATURES_POS][_POS_INVERSE if inverse else _POS_DIRECT]


import unittest
from shexer.shaper import Shaper
from test.const import G1, BASE_FILES, default_namespaces, G1_ALL_CLASSES_NO_COMMENTS
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth

from shexer.consts import TURTLE_ITER, NT



_BASE_DIR = BASE_FILES + "bnodes" + pth.sep  # We just need something with another instantiation property


class TestBNodeHandling(unittest.TestCase):

    def test_format_turtle_iter(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "bnode_people.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE_ITER,
            disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "schema_bnode_people.shex",
                                                      str_target=str_result,
                                                      strict=False))

    def test_format_nt(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "bnode_people.nt",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=NT,
            disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "schema_bnode_people.shex",
                                                      str_target=str_result,
                                                      strict=False))


    def test_some_bnodes_dont_have_shape(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "people_some_bnodes_dont_have_shape.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE_ITER,
            disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "schema_people_some_bnodes_dont_have_shape.shex",
                                                      str_target=str_result,
                                                      strict=False))

    def test_some_bnodes_dont_have_shape_comments(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "people_some_bnodes_dont_have_shape.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE_ITER,
            disable_comments=False)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue("100.0 % obj: BNode" in str_result)
        self.assertTrue("50.0 % obj: @:person" in str_result)

    def test_target_mixes_bnodes_and_iris(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "people_target_mixes_bnodes_and_iris.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE_ITER,
            disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(
            file_vs_str_tunned_comparison(file_path=_BASE_DIR + "schema_bnode_people.shex",
                                          str_target=str_result))

    def test_target_mixes_bnodes_and_iris_comments(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "people_target_mixes_bnodes_and_iris.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE_ITER,
            disable_comments=False,
            decimals=2)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue("66.67 % obj: @:person" in str_result)
        self.assertTrue("33.33 % obj: BNode" in str_result)
        self.assertTrue("33.33 % obj: IRI" in str_result)

    def test_target_mixes_bnodes_iris_and_no_shapes(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "people_target_mixes_bnodes_iris_and_no_shapes.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE_ITER,
            disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(
            file_vs_str_tunned_comparison(file_path=_BASE_DIR + "schema_people_target_mixes_bnodes_iris_and_no_shapes.shex",
                                          str_target=str_result))

    def test_target_mixes_bnodes_iris_and_no_shapes_comments(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "people_target_mixes_bnodes_iris_and_no_shapes.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE_ITER,
            disable_comments=False,
            decimals=2)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue("66.67 % obj: @:person" in str_result)
        self.assertTrue("66.67 % obj: BNode" in str_result)
        self.assertTrue("33.33 % obj: IRI" in str_result)

    def test_enable_or_but_it_is_not_useful(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "bnode_people.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE_ITER,
            disable_comments=True,
            disable_or_statements=False)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "schema_bnode_people.shex",
                                                      str_target=str_result))

    def test_enable_or_without_redundant(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "people_some_bnodes_dont_have_shape.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE_ITER,
            disable_or_statements=False,
            disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "schema_people_some_bnodes_dont_have_shape.shex",
                                                      str_target=str_result,
                                                      strict=False))

    def test_enable_or_with_redundant_only_bnodes(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "people_some_bnodes_dont_have_shape.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE_ITER,
            disable_or_statements=False,
            allow_redundant_or=True,
            disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "or_with_redundant_bnodes_and_shapes.shex",
                                                      str_target=str_result,
                                                      strict=False))


    def test_enable_or_without_redundant_with_shapes_bnodes_and_iris(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "people_target_mixes_bnodes_iris_and_no_shapes.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE_ITER,
            disable_comments=True,
            disable_or_statements=False)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(
            file_vs_str_tunned_comparison(file_path=_BASE_DIR + "schema_people_target_mixes_bnodes_iris_and_no_shapes.shex",
                                          str_target=str_result))

    def test_enable_or_with_redundant_with_shapes_bnodes_and_iris(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "people_target_mixes_bnodes_iris_and_no_shapes.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE_ITER,
            disable_comments=True,
            disable_or_statements=False,
            allow_redundant_or=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(
            file_vs_str_tunned_comparison(file_path=_BASE_DIR + "or_with_redundant_bnodes_iris_and_shapes.shex",
                                          str_target=str_result))

    def test_enable_or_with_redundant_with_shapes_bnodes_and_iris_comments(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "people_target_mixes_bnodes_iris_and_no_shapes.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE_ITER,
            disable_comments=False,
            decimals=2,
            disable_or_statements=False,
            allow_redundant_or=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue("66.67 % obj: @:person" in str_result)
        self.assertTrue("66.67 % obj: BNode" in str_result)
        self.assertTrue("33.33 % obj: IRI" in str_result)






from shexer.model.shape import STARTING_CHAR_FOR_SHAPE_NAME
from shexer.utils.uri import prefixize_uri_if_possible, get_prefix_of_namespace_if_it_exists
from shexer.io.shex.formater.consts import SHAPE_LINK_CHAR

def _candidate_name_for_class_uri(class_uri, shapes_namespace):
    if class_uri.startswith(SHAPE_LINK_CHAR): # special shape case
        return class_uri
    if class_uri.startswith("<") and class_uri.endswith(">"):
        return STARTING_CHAR_FOR_SHAPE_NAME + class_uri
    last_piece = class_uri
    if "#" in last_piece and last_piece[-1] != "#":
        last_piece = last_piece[last_piece.rfind("#") + 1:]
    if "/" in last_piece:
        if last_piece[-1] != "/":
            last_piece = last_piece[last_piece.rfind("/") + 1:]
        else:
            last_piece = last_piece[last_piece[:-1].rfind("/") + 1:]
    if last_piece.endswith(">"):
        last_piece = last_piece[:-1]
    if last_piece.startswith("<"):
        last_piece = last_piece[1:]
    return (shapes_namespace + last_piece) if last_piece is not None else class_uri
    # return STARTING_CHAR_FOR_SHAPE_NAME + "<" + shapes_namespace + last_piece + ">" if last_piece is not None else class_uri

def _refine_candidate_shape_name(candidate, class_uri, namespace_prefix_dict):
    prefix = get_prefix_of_namespace_if_it_exists(target_uri=class_uri,
                                                  corners=False,
                                                  namespaces_prefix_dict=namespace_prefix_dict)
    if prefix is None:
        return candidate
    return f"{candidate}_{prefix}"


def build_shapes_name_for_class_uri(class_uri, shapes_namespace, shape_names_dict, namespace_prefix_dict):
    candidate = _candidate_name_for_class_uri(class_uri=class_uri,
                                              shapes_namespace=shapes_namespace)
    names = shape_names_dict.values()
    if candidate not in names:
        return candidate
    candidate = _refine_candidate_shape_name(candidate=candidate,
                                             class_uri=class_uri,
                                             namespace_prefix_dict=namespace_prefix_dict)
    if candidate not in names:
        return candidate
    counter = 2
    original = candidate
    while candidate in names:
        candidate = original + str(counter)
        counter += 1
    return candidate






def build_shape_name_for_qualifier_prop_uri(prop_uri, shapes_namespace):  # TODO REVIEW!
    last_piece = None
    if "#" in prop_uri and prop_uri[-1] != "#":
        last_piece = prop_uri[prop_uri.rfind("#") + 1:]
    if "/" in prop_uri:
        if prop_uri[-1] != "/":
            last_piece = prop_uri[prop_uri.rfind("/") + 1:]
        else:
            last_piece = prop_uri[prop_uri[:-1].rfind("/") + 1:]
    if last_piece is not None:
        return STARTING_CHAR_FOR_SHAPE_NAME + "<" + shapes_namespace + last_piece + ">"
    return STARTING_CHAR_FOR_SHAPE_NAME + prop_uri.upper()


def prefixize_shape_name_if_possible(a_shape_name, namespaces_prefix_dict):
    result = prefixize_uri_if_possible(target_uri=a_shape_name[1:],                  # Avoid the "from shexer.model.shape. STARTING_CHAR_FOR_SHAPE_NAME starting char
                                       namespaces_prefix_dict=namespaces_prefix_dict)
    return result


from shexer.model.const_elem_types import IRI_ELEM_TYPE


class IRI(object):

    def __init__(self, content):
        self._content = content

    def __str__(self):
        return self._content

    @property
    def elem_type(self):
        return IRI_ELEM_TYPE

    @property
    def iri(self):
        return self._content

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        return str(self) == str(other)

    def __ne__(self, other):
        return not self.__eq__(other)


import unittest
from shexer.shaper import Shaper
from test.const import G1_NT, G1, BASE_FILES, default_namespaces, G1_ALL_CLASSES_NO_COMMENTS
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth
from shexer.consts import TURTLE, NT

_BASE_DIR = BASE_FILES + "namespaces_dict" + pth.sep

class TestNamespacesDict(unittest.TestCase):

    def test_same_namespaces_as_source_ttl_file(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_no_foaf(self):
        namespaces = {"http://example.org/" : "ex",
                               "http://www.w3.org/XML/1998/namespace/" : "xml",
                               "http://www.w3.org/1999/02/22-rdf-syntax-ns#": "rdf",
                               "http://www.w3.org/2000/01/rdf-schema#" : "rdfs",
                               "http://www.w3.org/2001/XMLSchema#": "xsd"
                               # "http://xmlns.com/foaf/0.1/": "foaf"
                               }
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1_NT,
                        namespaces_dict=namespaces,
                        all_classes_mode=False,
                        input_format=NT,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + pth.sep +"no_foaf.shex",
                                                      str_target=str_result))

    def test_overwrite_empty(self):
        namespaces = default_namespaces()
        namespaces["http://unuseful.but.yet/here/"] = ""

        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1,
                        namespaces_dict=namespaces,
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + pth.sep +"overwrite_empty.shex",
                                                      str_target=str_result))

    def test_overwrite_some_namespaces(self):
        namespaces = {"http://example.org/": "ex",
                      "http://www.w3.org/XML/1998/namespace/": "xml",
                      "http://www.w3.org/1999/02/22-rdf-syntax-ns#": "rdf",
                      "http://www.w3.org/2000/01/rdf-schema#": "rdfs",
                      "http://www.w3.org/2001/XMLSchema#": "xxssdd",
                      "http://xmlns.com/foaf/0.1/": "fooo"
                      }

        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1,
                        namespaces_dict=namespaces,
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + pth.sep +"overwrite.shex",
                                                      str_target=str_result))



from shexer.core.shexing.class_shexer import ClassShexer
from shexer.core.shexing.class_shexer_fed_sources import ClassShexerFedSources
from shexer.consts import RDF_TYPE, SHAPES_DEFAULT_NAMESPACE, RATIO_INSTANCES, FREQ_PROP, ABSOLUTE_COUNT_PROP, \
    EXTRA_INFO_PROP
from shexer.utils.target_elements import tune_target_classes_if_needed
from shexer.utils.dict import reverse_keys_and_values


def get_class_shexer(class_counts,
                     class_profile_dict,
                     remove_empty_shapes=True,
                     original_target_classes=None,
                     original_shape_map=None,
                     discard_useless_constraints_with_positive_closure=True,
                     keep_less_specific=True,
                     all_compliant_mode=True,
                     instantiation_property=RDF_TYPE,
                     disable_or_statements=True,
                     disable_comments=False,
                     namespaces_dict=None,
                     allow_opt_cardinality=True,
                     disable_exact_cardinality=False,
                     shapes_namespace=SHAPES_DEFAULT_NAMESPACE,
                     inverse_paths=False,
                     decimals=-1,
                     instances_report_mode=RATIO_INSTANCES,
                     detect_minimal_iri=False,
                     class_min_iris=None,
                     allow_redundant_or=False,
                     federated_sources=None,
                     shape_names=None,
                     frequency_property=FREQ_PROP,
                     comments_to_annotations=False,
                     absolute_counts_prop=ABSOLUTE_COUNT_PROP,
                     extra_info_property=EXTRA_INFO_PROP):


    if original_target_classes is not None:
        original_target_classes = tune_target_classes_if_needed(original_target_classes, reverse_keys_and_values(namespaces_dict))

    if federated_sources is None:
        return ClassShexer(
            class_counts_dict=class_counts,
            class_profile_dict=class_profile_dict,
            class_profile_json_file=None,
            remove_empty_shapes=remove_empty_shapes,
            original_target_classes=original_target_classes,
            original_shape_map=original_shape_map,
            discard_useless_constraints_with_positive_closure=discard_useless_constraints_with_positive_closure,
            keep_less_specific=keep_less_specific,
            all_compliant_mode=all_compliant_mode,
            instantiation_property=instantiation_property,
            disable_or_statements=disable_or_statements,
            disable_comments=disable_comments,
            namespaces_dict=namespaces_dict,
            allow_opt_cardinality=allow_opt_cardinality,
            disable_exact_cardinality=disable_exact_cardinality,
            shapes_namespace=shapes_namespace,
            inverse_paths=inverse_paths,
            instances_report_mode=instances_report_mode,
            decimals=decimals,
            detect_minimal_iri=detect_minimal_iri,
            class_min_iris_dict=class_min_iris,
            allow_redundant_or=allow_redundant_or,
            shape_names_dict=shape_names,
            frequency_property=frequency_property,
            comments_to_annotations=comments_to_annotations,
            absolute_count_property=absolute_counts_prop,
            extra_info_property=extra_info_property
        )
    else:
        return ClassShexerFedSources(
            class_counts_dict=class_counts,
            class_profile_dict=class_profile_dict,
            class_profile_json_file=None,
            remove_empty_shapes=remove_empty_shapes,
            original_target_classes=original_target_classes,
            original_shape_map=original_shape_map,
            discard_useless_constraints_with_positive_closure=discard_useless_constraints_with_positive_closure,
            keep_less_specific=keep_less_specific,
            all_compliant_mode=all_compliant_mode,
            instantiation_property=instantiation_property,
            disable_or_statements=disable_or_statements,
            disable_comments=disable_comments,
            namespaces_dict=namespaces_dict,
            allow_opt_cardinality=allow_opt_cardinality,
            disable_exact_cardinality=disable_exact_cardinality,
            shapes_namespace=shapes_namespace,
            inverse_paths=inverse_paths,
            instances_report_mode=instances_report_mode,
            decimals=decimals,
            detect_minimal_iri=detect_minimal_iri,
            class_min_iris_dict=class_min_iris,
            allow_redundant_or=allow_redundant_or,
            fed_sources=federated_sources,
            shape_names_dict=shape_names,
            frequency_property=frequency_property,
            comments_to_annotations=comments_to_annotations,
            absolute_counts_prop=absolute_counts_prop,
            extra_info_property=extra_info_property
        )



from shexer.io.graph.yielder.base_triples_yielder import BaseTriplesYielder
from shexer.io.graph.yielder.multifile_base_triples_yielder import MultifileBaseTripleYielder
from shexer.utils.triple_yielders import tune_subj, tune_prop, tune_token
from shexer.utils.uri import remove_corners
import tempfile
from pathlib import Path
import re
from shexer.utils.exception import ParseError

WHITESPACES= re.compile(r"\s+")

import lightrdf


class LightTurtleTriplesYielder(BaseTriplesYielder):

    def __init__(self, source_file, raw_graph, namespaces_dict):
        super().__init__()
        self._prefixes = {}
        self._source_file = source_file
        self._raw_graph = raw_graph
        self._namespaces_dict = namespaces_dict if namespaces_dict is not None else {}
        self._yielded_triples = 0

    def _yield_triples(self):
        self._extract_prefixes()
        parser = lightrdf.turtle.Parser()
        try:
            for s, p, o in parser.parse(self._source_file, base_iri=None):
                yield (
                    tune_subj(s),
                    tune_prop(p),
                    tune_token(o)
                )
                self._yielded_triples += 1
        except BaseException as e:
            raise ParseError(f"Error while parsing: {e}") from e

    def yield_triples(self):
        if self._raw_graph is not None:
            with tempfile.TemporaryDirectory() as tmpdir:
                self._source_file = Path(tmpdir) / "data.nt"
                self._source_file.write_text(self._raw_graph, encoding="utf-8")
                self._source_file = str(self._source_file)
                for a_triple in self._yield_triples():
                    yield a_triple
        else:
            for a_triple in self._yield_triples():
                yield a_triple

    @property
    def yielded_triples(self):
        return self._yielded_triples

    @property
    def error_triples(self):  # No error triples in this parser, it crashes when finding an error
        return 0

    def _extract_prefixes(self):
        with open(self._source_file, "r", encoding="utf-8") as f:
            for line in f:
                line = line.strip()
                line = WHITESPACES.sub(" ", line)
                if line.startswith("@prefix"):
                    self._process_prefix_line(line)
                else:  # if declarations are not consecutive and at the beginning, it will not work
                    break

    def _process_prefix_line(self, line):
        pieces = line.split(" ")
        prefix = pieces[1] if not pieces[1].endswith(":") else pieces[1][: - 1]
        base_url = remove_corners(pieces[2])
        if base_url not in self._namespaces_dict:
            self._namespaces_dict[base_url] = prefix

class MultiLightTurtleTriplesYielder(MultifileBaseTripleYielder):
    def __init__(self, list_of_files, namespaces_dict):
        super(MultiLightTurtleTriplesYielder, self).__init__(
            list_of_files=list_of_files,
            namespaces_to_ignore=None,
            allow_untyped_numbers=False,
            compression_mode=None,
            zip_base_archive=None)
        self._namespaces_dict = namespaces_dict

    def _constructor_file_yielder(self, a_source_file, parse_namespaces=False):
        return LightTurtleTriplesYielder(source_file=a_source_file,
                                         namespaces_dict=self._namespaces_dict,
                                         raw_graph=None)




import unittest
from shexer.shaper import Shaper
from test.const import default_namespaces, BASE_FILES_GENERAL
from test.t_utils import file_vs_str_tunned_comparison

from shexer.consts import TURTLE

raw_graph_nt = """<http://example.org/Jimmy> <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <http://xmlns.com/foaf/0.1/Person> .
<http://example.org/Jimmy> <http://xmlns.com/foaf/0.1/age> "23"^^<http://www.w3.org/2001/XMLSchema#integer> .
<http://example.org/Jimmy> <http://xmlns.com/foaf/0.1/name> "Jimmy" .
<http://example.org/Jimmy> <http://xmlns.com/foaf/0.1/familyName> "Jones" .
<http://example.org/Sarah> <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <http://xmlns.com/foaf/0.1/Person> .
<http://example.org/Sarah> <http://xmlns.com/foaf/0.1/age> "22"^^<http://www.w3.org/2001/XMLSchema#integer> .
<http://example.org/Sarah> <http://xmlns.com/foaf/0.1/name> "Sarah" .
<http://example.org/Sarah> <http://xmlns.com/foaf/0.1/familyName> "Salem" .
<http://example.org/Bella> <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <http://xmlns.com/foaf/0.1/Person> .
<http://example.org/Bella> <http://xmlns.com/foaf/0.1/age> "56"^^<http://www.w3.org/2001/XMLSchema#integer> .
<http://example.org/Bella> <http://xmlns.com/foaf/0.1/name> "Isabella" .
<http://example.org/David> <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <http://xmlns.com/foaf/0.1/Person> .
<http://example.org/David> <http://xmlns.com/foaf/0.1/name> "David" .
<http://example.org/David> <http://xmlns.com/foaf/0.1/familyName> "Doulofeau" .
<http://example.org/David> <http://xmlns.com/foaf/0.1/knows> <http://example.org/Sarah> .
<http://example.org/HumanLike> <http://xmlns.com/foaf/0.1/name> "Person" .
<http://example.org/HumanLike> <http://xmlns.com/foaf/0.1/familyName> "Maybe" .
<http://example.org/HumanLike> <http://xmlns.com/foaf/0.1/age> "99"^^<http://www.w3.org/2001/XMLSchema#integer> .
<http://example.org/HumanLike> <http://xmlns.com/foaf/0.1/knows> <http://example.org/David> .
<http://example.org/x1> <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <http://xmlns.com/foaf/0.1/Document> .
<http://example.org/x1> <http://xmlns.com/foaf/0.1/depiction> "A thing that s nice" .
<http://example.org/x1> <http://xmlns.com/foaf/0.1/title> "A nice thing" .
<http://example.org/x2> <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <http://xmlns.com/foaf/0.1/Document> .
<http://example.org/x2> <http://xmlns.com/foaf/0.1/title> "Another thing" ."""

raw_graph_turtle = """@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix ex: <http://example.org/> .
@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

ex:Jimmy a foaf:Person ;  # Complete
	foaf:age "23"^^xsd:integer ;
	foaf:name "Jimmy" ;
	foaf:familyName "Jones" .

ex:Sarah a foaf:Person ;  # Complete implicit type for age
	foaf:age 22 ;
	foaf:name "Sarah" ;
	foaf:familyName "Salem" .

ex:Bella a foaf:Person ;  # Missing familyName
	foaf:age "56"^^xsd:integer ;
	foaf:name "Isabella" .

ex:David a foaf:Person ;  # Missing age and use knows
	foaf:name "David" ;
	foaf:familyName "Doulofeau" ;
	foaf:knows ex:Sarah .

ex:HumanLike foaf:name "Person" ;  # foaf properties, but not explicit type.
	foaf:familyName "Maybe" ;
	foaf:age 99 ;
	foaf:knows ex:David .


ex:x1 rdf:type foaf:Document ;
	foaf:depiction "A thing that is nice" ;
	foaf:title "A nice thing" .


ex:x2 rdf:type foaf:Document ;
	foaf:title "Another thing" ."""


class TestGraphFileInput(unittest.TestCase):

    def test_some_format(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        raw_graph=raw_graph_turtle,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=BASE_FILES_GENERAL + "g1_all_classes_no_comments.shex",
                                                      str_target=str_result,
                                                      strict=False))


    def test_no_format(self):  # Should be nt
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        raw_graph=raw_graph_nt,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=BASE_FILES_GENERAL + "g1_all_classes_no_comments.shex",
                                                      str_target=str_result))


from shexer.model.IRI import IRI
from shexer.model.property import Property
from shexer.model.Literal import Literal
from shexer.model.bnode import BNode
from shexer.utils.uri import remove_corners
from shexer.utils.literal import parse_literal, parse_unquoted_literal, FLOAT_TYPE, INTEGER_TYPE


def check_if_property_belongs_to_namespace_list(str_prop, namespaces):
    """
    It return True if the property balongs to some namespace directly, i.e.,
    without adding any hierarchical element before reaching the name of the property itself.
    Example:
    Property http:example.org/prop, namespace http:example.org/ ---> True
    Property http:example.org/properties/prop, namespace http:example.org/ ---> False
    :param str_prop:
    :param namespaces:
    :return:
    """
    for a_namespace in namespaces:
        if str_prop.startswith(a_namespace):
            if "/" not in str_prop[len(a_namespace):] and "#" not in str_prop[len(a_namespace):]:
                return True
    return False


def tune_subj(a_token, raise_error_if_no_corners=True):
    if a_token.startswith("<"):
        return IRI(remove_corners(a_uri=a_token,
                                  raise_error_if_no_corners=raise_error_if_no_corners))
    elif a_token.startswith("_:"):
        return BNode(identifier=a_token)
    elif a_token.strip() == "[]":
        return BNode(identifier=a_token)

    else:  # ???
        raise ValueError("Unrecognized token in subject position: " + a_token)


def tune_token(a_token, allow_untyped_numbers=False, raise_error_if_no_corners=True, base_namespace=None):
    if a_token.startswith("<"):
        return IRI(remove_corners(a_uri=a_token,
                                  raise_error_if_no_corners=raise_error_if_no_corners))
    elif a_token.startswith('"'):
        content, elem_type = parse_literal(an_elem=a_token,
                                           base_namespace=base_namespace)
        return Literal(content=content,
                       elem_type=elem_type)
    elif a_token.startswith("_:"):
        return BNode(identifier=a_token)
    elif a_token.strip() == "[]":
        return BNode(identifier=a_token)
    if allow_untyped_numbers:
        try:
            candidate_float = float(a_token)
            if _is_integer(candidate_float):
                return Literal(content=a_token.strip(),
                               elem_type=INTEGER_TYPE)
            return Literal(content=a_token.strip(),
                           elem_type=FLOAT_TYPE)
        except:
            pass

    content, elem_type = parse_unquoted_literal(a_token)
    return Literal(content=content,
                   elem_type=elem_type)


def _is_integer(float_number):
    if float_number % 1.0 == 0:
        return True
    return False


def tune_prop(a_token, raise_error_if_no_corners=True):
    return Property(remove_corners(a_uri=a_token,
                                   raise_error_if_no_corners=raise_error_if_no_corners))



import unittest
from shexer.shaper import Shaper
from test.const import G1, BASE_FILES, default_namespaces, G1_ALL_CLASSES_NO_COMMENTS
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth

from shexer.consts import TURTLE_ITER, GZ, ZIP, XZ, N3, TURTLE, RDF_XML, TSV_SPO, NT, JSON_LD



_BASE_DIR = BASE_FILES + "compression" + pth.sep


class TestCompressionMode(unittest.TestCase):

    ######################### GZ

    def test_ttl_iter_gz(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.ttl.gz",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE_ITER,
            disable_comments=True,
            compression_mode=GZ
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_n3_gz(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.n3.gz",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=N3,
            disable_comments=True,
            compression_mode=GZ
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_json_gz(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.json.gz",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=JSON_LD,
            disable_comments=True,
            compression_mode=GZ
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_ttl_rdflib_gz(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.ttl.gz",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=True,
            compression_mode=GZ
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_xml_gz(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.xml.gz",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=RDF_XML,
            disable_comments=True,
            compression_mode=GZ
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_tsv_spo_gz(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.tsv.gz",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TSV_SPO,
            disable_comments=True,
            compression_mode=GZ
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_nt_gz(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.nt.gz",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=NT,
            disable_comments=True,
            compression_mode=GZ
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    ############# zip

    def test_ttl_iter_zip(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.ttl.zip",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE_ITER,
            disable_comments=True,
            compression_mode=ZIP
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_n3_zip(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.n3.zip",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=N3,
            disable_comments=True,
            compression_mode=ZIP
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_ttl_rdflib_zip(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.ttl.zip",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=True,
            compression_mode=ZIP
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_xml_zip(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.xml.zip",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=RDF_XML,
            disable_comments=True,
            compression_mode=ZIP
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_tsv_spo_zip(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.tsv.zip",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TSV_SPO,
            disable_comments=True,
            compression_mode=ZIP
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_nt_zip(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.nt.zip",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=NT,
            disable_comments=True,
            compression_mode=ZIP
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_json_zip(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.json.zip",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=JSON_LD,
            disable_comments=True,
            compression_mode=ZIP
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))


    ########################  xz

    def test_ttl_iter_xz(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.ttl.xz",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE_ITER,
            disable_comments=True,
            compression_mode=XZ
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_n3_xz(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.n3.xz",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=N3,
            disable_comments=True,
            compression_mode=XZ
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_json_xz(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.json.xz",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=JSON_LD,
            disable_comments=True,
            compression_mode=XZ
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_ttl_rdflib_xz(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.ttl.xz",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=True,
            compression_mode=XZ
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_xml_xz(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.xml.xz",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=RDF_XML,
            disable_comments=True,
            compression_mode=XZ
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_tsv_spo_xz(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.tsv.xz",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TSV_SPO,
            disable_comments=True,
            compression_mode=XZ
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_nt_xz(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t_graph_1.nt.xz",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=NT,
            disable_comments=True,
            compression_mode=XZ
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    ########################  Wrong params

    def test_unknown_mode(self):
        try:
            shaper = Shaper(
                graph_file_input=_BASE_DIR + "t_graph_1.json.zip",
                namespaces_dict=default_namespaces(),
                all_classes_mode=True,
                input_format=JSON_LD,
                disable_comments=True,
                compression_mode="RAR"
            )
            self.fail("It shouldn`t allow to use an unknown compression format")
        except ValueError:
            pass  # thats ok
        except:
            self.fail("The exception should be a ValueError")

    def test_remote_source(self):
        shape_map_raw = "SPARQL'select ?p where " \
                        "{ ?p <http://www.wikidata.org/prop/direct/P31> <http://www.wikidata.org/entity/Q14660> } " \
                        "LIMIT 1'@<Flag>"
        try:
            shaper = Shaper(shape_map_raw=shape_map_raw,
                            url_endpoint="https://query.wikidata.org/sparql",
                            namespaces_dict=default_namespaces(),
                            instantiation_property="http://www.wikidata.org/prop/direct/P31",
                            disable_comments=True,
                            depth_for_building_subgraph=1,
                            track_classes_for_entities_at_last_depth_level=False,
                            all_classes_mode=False,
                            compression_mode=ZIP)
            self.fail("It should allow to use compression with a remote source")
        except ValueError:
            pass
        except:
            self.fail("The exception should be a ValueError")


    ##################### SEVERAL FILES IN ZIP

    def test_nt_zip(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "t1_graph_partials_1_2_3__3.nt.zip",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=NT,
            disable_comments=True,
            compression_mode=ZIP
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    ##################### SEVERAL ZIPS

    def test_nt_zip(self):
        shaper = Shaper(
            graph_list_of_files_input=[_BASE_DIR + "t1_graph_partials_1_2__3.nt.zip",
                                       _BASE_DIR + "t1_graph_partials_3__3.nt.zip"],
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=NT,
            disable_comments=True,
            compression_mode=ZIP
        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))



from shexer.io.shex.formater.statement_serializers.base_statement_serializer import BaseStatementSerializer, \
    SPACES_GAP_BETWEEN_TOKENS

#
# class InverseStatementSerializer(BaseStatementSerializer):
#
#     def __init__(self, ref_statement_serializer):
#         super().__init__(ref_statement_serializer._instantiation_property_str)
#         self._ref_statement_serializer = ref_statement_serializer
#
#     def serialize_statement_with_indent_level(self, a_statement, is_last_statement_of_shape, namespaces_dict):
#         base_result = super().serialize_statement_with_indent_level(
#             a_statement=a_statement,
#             is_last_statement_of_shape=is_last_statement_of_shape,
#             namespaces_dict=namespaces_dict)
#         if len(base_result) == 0:
#             return base_result
#         self._add_inverse_sense_to_first_tuple(base_result)
#         return base_result
#
#     def _add_inverse_sense_to_first_tuple(self, statement_str_indent_tuples):
#         """
#         This method modifies the input, no return needed.
#         :param statement_str_indent_tuples:
#         :return:
#         """
#         statement_str_indent_tuples[0][0] = _INVERSE_SENSE_SHEXC + \
#                                             SPACES_GAP_BETWEEN_TOKENS + \
#                                             statement_str_indent_tuples[0][0]
#
#     def _sense_flag(self):
#         return _INVERSE_SENSE_SHEXC + SPACES_GAP_BETWEEN_TOKENS



class Annotation(object):

    def __init__(self, predicate, obj):
        self._predicate = predicate
        self._obj = obj

    @property
    def predicate(self):
        return self._predicate

    @predicate.setter
    def predicate(self, new_pred):
        self._predicate = new_pred

    @property
    def obj(self):
        return self._obj

    @obj.setter
    def obj(self, new_obj):
        self._obj = new_obj




from shexer.model.hierarchy_tree import HTree, HNode, HMacro
from shexer.model.Macro import Macro
from shexer.model.const_elem_types import DOT_ELEM_TYPE, IRI_ELEM_TYPE, LITERAL_ELEM_TYPE


def get_basic_h_tree():
    result = HTree()
    dot_node = HNode(hcontent=HMacro(value=Macro(macro_const=DOT_ELEM_TYPE)),
                     htree=result)
    result.root = dot_node
    literal_node = HNode(hcontent=HMacro(value=Macro(macro_const=LITERAL_ELEM_TYPE)),
                         htree=result)
    iri_node = HNode(hcontent=HMacro(value=Macro(macro_const=IRI_ELEM_TYPE)),
                     htree=result)

    dot_node.add_child(literal_node)
    dot_node.add_child(iri_node)
    return result






#OUTPUTS
SHEXC = "ShEx"
SHACL_TURTLE = "Shacl"

#INPUT FORMATS
NT = "nt"
TSV_SPO = "tsv_spo"
TURTLE = "turtle"
TURTLE_ITER = "turtle_iter"
RDF_XML = "xml"
N3 = "n3"
JSON_LD = "json-ld"

#SHAPE MAP FORMATS
JSON = "json"
FIXED_SHAPE_MAP = "fsm"

#FREQUENT INSTATIATION PROPERTIES
RDF_TYPE = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"
WIKIDATA_INSTACE_OF = "http://www.wikidata.org/prop/direct/P31"

#NAMESPACES
SHAPES_DEFAULT_NAMESPACE = "http://weso.es/shapes/"

#COMPRESSION FORMATS
ZIP = "zip"
GZ = "gz"
XZ = "xz"

# FREQUENCY MODES

RATIO_INSTANCES = "ratio"
ABSOLUTE_INSTANCES = "abs"
MIXED_INSTANCES = "mixed"


# EXAMPLES
SHAPE_EXAMPLES = "shape"
CONSTRAINT_EXAMPLES = "cons"
ALL_EXAMPLES = "all"

# UML
UML_PLANT_SERVER = "http://www.plantuml.com/plantuml/img/"

DEFAULT_NAMESPACES = {"http://example.org/": "ex",
            "http://www.w3.org/XML/1998/namespace/": "xml",
            "http://www.w3.org/1999/02/22-rdf-syntax-ns#": "rdf",
            "http://www.w3.org/2000/01/rdf-schema#": "rdfs",
            "http://www.w3.org/2001/XMLSchema#": "xsd",
            "http://xmlns.com/foaf/0.1/": "foaf"
            }

# WESO-SHAPES-ONTO  # todo !
FREQ_PROP = "http://weso.es/shexer/ontology/ratio_property_usage"
EXTRA_INFO_PROP = "http://www.w3.org/2000/01/rdf-schema#comment"
EXAMPLE_CONFORMANCE_PROP = "http://weso.es/shexer/ontology/conformant_example"
ABSOLUTE_COUNT_PROP = "http://weso.es/shexer/ontology/total_conforming_instances"


import unittest
from shexer.shaper import Shaper
from test.const import BASE_FILES, default_namespaces, G1_ALL_CLASSES_NO_COMMENTS
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth

from shexer.consts import TURTLE



_BASE_DIR = BASE_FILES + "repeated_names" + pth.sep  # We just need something with another instantiation property


class TestRepeatedShapeNames(unittest.TestCase):

    def test_potentially_repeated_shape_name(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "two_person_classes.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "two_person_classes.shex",
                                                      str_target=str_result)
                        or
                        file_vs_str_tunned_comparison(file_path=_BASE_DIR + "two_person_classes_op2.shex",
                                                      str_target=str_result)
                        )  # There is a non-deterministic part. Different shape names could be generated
                           # depending on the first shape explored




from shexer.io.shex.formater.statement_serializers.frequency_strategy.base_frequency_strategy import \
    BaseFrequencyStrategy
from shexer.io.shex.formater.consts import SPACES_GAP_BETWEEN_TOKENS, ANNOTATION_BEGIN
from shexer.utils.uri import prefixize_uri_if_possible, add_corners_if_it_is_an_uri


class AbsFreqSerializer(BaseFrequencyStrategy):

    def __init__(self, frequency_ratio_property, frequency_absolute_property, namespaces_dict):
        super().__init__(frequency_ratio_property, frequency_absolute_property, namespaces_dict)

    def serialize_frequency(self, statement):
        return str(statement.n_occurences) + " instance{}.".format(
            "" if statement.n_occurences == 1
            else "s"
        )

    def annotations_for_frequency(self, statement):
        freq_annotation = SPACES_GAP_BETWEEN_TOKENS.join((ANNOTATION_BEGIN,
                                                          add_corners_if_it_is_an_uri(prefixize_uri_if_possible(
                                                              target_uri=self._frequency_absolute_property,
                                                              namespaces_prefix_dict=self._namespaces_dict,
                                                              corners=False)),
                                                          str(statement.n_occurences)
                                                          ))
        return [freq_annotation]



from shexer.io.shex.formater.statement_serializers.base_statement_serializer import BaseStatementSerializer
from shexer.io.shex.formater.consts import SPACES_GAP_BETWEEN_TOKENS, KLEENE_CLOSURE, OPT_CARDINALITY
from shexer.consts import FREQ_PROP, ABSOLUTE_COUNT_PROP, EXTRA_INFO_PROP


class FixedPropChoiceStatementSerializer(BaseStatementSerializer):

    def __init__(self, instantiation_property_str, frequency_serializer, disable_comments=False, is_inverse=False,
                 frequency_property=FREQ_PROP, namespaces_dict=None, comments_to_annotations=False,
                 extra_info_prop=EXTRA_INFO_PROP, absolute_count_prop=ABSOLUTE_COUNT_PROP):
        super(FixedPropChoiceStatementSerializer, self).__init__(instantiation_property_str=instantiation_property_str,
                                                                 disable_comments=disable_comments,
                                                                 is_inverse=is_inverse,
                                                                 frequency_serializer=frequency_serializer,
                                                                 frequency_property=frequency_property,
                                                                 namespaces_dict=namespaces_dict,
                                                                 comments_to_annotations=comments_to_annotations,
                                                                 extra_info_prop=extra_info_prop,
                                                                 absolute_count_prop=absolute_count_prop)

    def serialize_statement_with_indent_level(self, a_statement, is_last_statement_of_shape):
        tuples_line_indent = []
        st_property = BaseStatementSerializer.tune_token(a_statement.st_property, self._namespaces_dict)
        st_target_elements = []
        for a_type in a_statement.st_types:
            st_target_elements.append(self.str_of_target_element(target_element=a_type,
                                                                 st_property=a_statement.st_property))

        content_line = st_property + SPACES_GAP_BETWEEN_TOKENS
        content_line += (SPACES_GAP_BETWEEN_TOKENS + "OR" + SPACES_GAP_BETWEEN_TOKENS).join(st_target_elements)
        content_line += SPACES_GAP_BETWEEN_TOKENS + BaseStatementSerializer.cardinality_representation(
            statement=a_statement,
            out_of_comment=True)
        if self._comments_to_annotations:
            content_line += self._build_constraint_annotations(a_statement)
        content_line += ";" if not is_last_statement_of_shape else ""
        tuples_line_indent.append((content_line, 1))

        if not self._comments_to_annotations:
            for a_comment in a_statement.comments:
                tuples_line_indent.append((a_comment, 4))
        return tuples_line_indent


    @staticmethod
    def turn_statement_into_comment(statement, namespaces_dict):
        return statement.probability_representation() + \
               " with cardinality " + statement.cardinality_representation()



from shexer.io.graph.yielder.multifile_base_triples_yielder import MultifileBaseTripleYielder
from shexer.io.graph.yielder.rdflib_triple_yielder import RdflibParserTripleYielder
from shexer.consts import TURTLE


class MultiRdfLibTripleYielder(MultifileBaseTripleYielder):

    def __init__(self, list_of_files, input_format=TURTLE, allow_untyped_numbers=False,
                 namespaces_dict=None, compression_mode=None, zip_archive_file=None):
        super(MultiRdfLibTripleYielder, self).__init__(list_of_files=list_of_files,
                                                       allow_untyped_numbers=allow_untyped_numbers)

        self._input_format = input_format
        self._namespaces_dict = namespaces_dict if namespaces_dict is not None else {}
        self._compression_mode = compression_mode
        self._zip_archive_file = zip_archive_file

    def _yield_triples_of_last_yielder(self, parse_namespaces=True):
        for a_triple in self._last_yielder.yield_triples(parse_namespaces):
            yield a_triple

    def _constructor_file_yielder(self, a_source_file):
        return RdflibParserTripleYielder(source=a_source_file,
                                         allow_untyped_numbers=self._allow_untyped_numbers,
                                         input_format=self._input_format,
                                         compression_mode=self._compression_mode,
                                         zip_archive_file=self._zip_archive_file)

    @property
    def namespaces(self):
        return self._namespaces_dict  # TODO This is not entirely correct. But this method will be rarely used
        # and can have a huge performance cost in case the graphs hadnt been parsed yet



from shexer.core.instances.abstract_instance_tracker import  _RDF_TYPE, _RDFS_SUBCLASS_OF
from shexer.core.instances.instance_tracker import InstanceTracker
from shexer.consts import SHAPES_DEFAULT_NAMESPACE
from shexer.model.bnode import BNode
from shexer.core.instances.pconsts import _S


class EndpointInstanceTracker(InstanceTracker):

    def __init__(self, target_classes, triples_yielder, instantiation_property=_RDF_TYPE, all_classes_mode=False,
                 subclass_property=_RDFS_SUBCLASS_OF, track_hierarchies=True, shape_qualifiers_mode=False,
                 namespaces_for_qualifier_props=None, shapes_namespace=SHAPES_DEFAULT_NAMESPACE, instances_cap=-1):
        super().__init__(target_classes=target_classes,
                         triples_yielder=triples_yielder,
                         instantiation_property=instantiation_property,
                         all_classes_mode=all_classes_mode,
                         subclass_property=subclass_property,
                         track_hierarchies=track_hierarchies,
                         shape_qualifiers_mode=shape_qualifiers_mode,
                         namespaces_for_qualifier_props=namespaces_for_qualifier_props,
                         shapes_namespace=shapes_namespace,
                         instances_cap=instances_cap)

    def _yield_relevant_triples(self):
        for a_triple in self._triples_yielder.yield_triples():
            if self._annotator.is_relevant_triple(a_triple) and self._subject_is_not_bnode(a_triple):
                self._relevant_triples += 1
                yield a_triple
            else:
                self._not_relevant_triples += 1

    def _subject_is_not_bnode(self, a_triple):
        return not isinstance(a_triple[_S], BNode)




_STR_BOOLS = ["true", "false"]
_STR_TEMPLATE = '"{}"'

def _add_prefix(unprefixed_elem, prefix):
    return prefix + ":" + unprefixed_elem


def remove_corners(a_uri, raise_error_if_no_corners=True):
    if a_uri.startswith("<") and a_uri.endswith(">"):
        return a_uri[1:-1]
    if raise_error_if_no_corners:
        raise ValueError("Wrong parameter of function: '" + a_uri + "'")
    else:
        return a_uri


def add_corners(a_uri):
    return "<" + a_uri + ">"

def add_corners_if_needed(a_uri):
    if a_uri.startswith("<"):
        return a_uri
    return add_corners(a_uri)

def longest_common_prefix(uri1, uri2):
    """
    It returns an str containing the longest possible common initial part of uri1 and uri2

    :param uri1:
    :param uri2:

    :return:
    """

    if len(uri1) == 0 or len(uri2) == 0:
        return ""
    shortest = len(uri1) if len(uri1) < len(uri2) else len(uri2)
    for i in range(shortest):
        if uri1[i] != uri2[i]:
            return uri1[:i]
    return uri1[:shortest]

def add_corners_if_it_is_an_uri(a_candidate_uri):
    if a_candidate_uri.startswith("http://") or a_candidate_uri.startswith("https://"):  # TODO, check this!
        return "<" + a_candidate_uri + ">"
    return a_candidate_uri


def is_a_correct_uri(target_uri, prefix_namespace_dict):
    """
    TODO: Here I am assuming that there is no forbiden char ( " < > # % { } | \ ^ ~ [ ] ` )
    :param target_uri:
    :param prefix_namespace_dict:
    :return:
    """
    if target_uri[0] == "<" and target_uri[-1] == ">":
        return True
    for a_prefix in prefix_namespace_dict:
        if target_uri.startswith(a_prefix + ":"):
            return True
        return False

def unprefixize_uri_if_possible(target_uri, prefix_namespaces_dict, include_corners=True):
    for a_prefix in prefix_namespaces_dict:
        if target_uri.startswith(a_prefix+":"):
            result = target_uri.replace(a_prefix+":", prefix_namespaces_dict[a_prefix])
            if include_corners:
                result = add_corners(result)
            return result
    return target_uri

def unprefixize_uri_mandatory(target_uri, prefix_namespaces_dict, include_corners=True):
    for a_prefix in prefix_namespaces_dict:
        if target_uri.startswith(a_prefix+":"):
            result = target_uri.replace(a_prefix+":", prefix_namespaces_dict[a_prefix])
            if include_corners:
                result = add_corners(result)
            return result
    raise ValueError("Unrecognized prefix in the following element" + target_uri)


def prefixize_uri_if_possible(target_uri, namespaces_prefix_dict, corners=True):
    best_match = None
    candidate_uri = remove_corners(target_uri) if corners else target_uri
    for a_namespace in namespaces_prefix_dict:  # Prefixed element (all literals are prefixed elements)
        if candidate_uri.startswith(a_namespace):
            if "/" not in candidate_uri[len(a_namespace):] and \
                "#" not in candidate_uri[len(a_namespace):]:
                best_match = a_namespace
                break
            # if best_match is None or len(best_match) < len(a_namespace):
            #     best_match = a_namespace

    return target_uri if best_match is None else candidate_uri.replace(best_match, namespaces_prefix_dict[best_match] + ":")


def get_prefix_of_namespace_if_it_exists(target_uri, namespaces_prefix_dict, corners=True):
    candidate_uri = remove_corners(target_uri) if corners else target_uri
    for a_namespace in namespaces_prefix_dict:  # Prefixed element (all literals are prefixed elements)
        if candidate_uri.startswith(a_namespace):
            if "/" not in candidate_uri[len(a_namespace):] and \
                    "#" not in candidate_uri[len(a_namespace):]:
                return namespaces_prefix_dict[a_namespace]


def serialize_as_triple_obj(sequence, namespaces_dict, reverse_namespaces_dict):
    if sequence.startswith("http"):
        return add_corners_if_needed(
            prefixize_uri_if_possible(target_uri=sequence,
                                      namespaces_prefix_dict=namespaces_dict,
                                      corners=False))

    elif sequence.startswith("<http"):
        return prefixize_uri_if_possible(target_uri=sequence,
                                         namespaces_prefix_dict=namespaces_dict,
                                         corners=True)
    # Case 1 --> a literal number
    # Case 2 --> bool keyword
    # Case 3 --> prefixed uri
    # In either case it goes as it is when serialized in a triple obj
    elif sequence.isnumeric() or \
            sequence in _STR_BOOLS or \
            is_a_correct_uri(target_uri=sequence, prefix_namespace_dict=reverse_namespaces_dict):
        return sequence
    else:  # Must be a string
        return _STR_TEMPLATE.format(sequence)




from shexer.utils.shapes import build_shapes_name_for_class_uri
from shexer.core.profiling.strategy.abstract_feature_direction_strategy import AbstractFeatureDirectionStrategy
from shexer.core.profiling.consts import _S, _P, _O, POS_CLASSES



class DirectFeaturesStrategy(AbstractFeatureDirectionStrategy):

    def __init__(self, class_profiler):
        super().__init__(class_profiler)
        self._set_annotation_methods()


    def adapt_instances_dict(self):
        for a_subj_key in self._i_dict:
            self._i_dict[a_subj_key] = \
                (self._i_dict[a_subj_key], {})

    def is_a_relevant_triple(self, a_triple):
        return self._is_relevant_instance(a_triple[_S])

    def _annotate_triple_features(self, a_triple):
        raise NotImplementedError()


    def annotate_instance_features(self, an_instance):
        self._annotate_direct_instance_features(an_instance)


    def init_annotated_targets(self):
        self._init_annotated_direct_features()

    def init_original_targets(self):
        if self._original_raw_target_classes:
            for a_class in self._original_raw_target_classes:
                self._c_shapes_dict[a_class] = {}
                self._c_counts[a_class] = 0
                self._shape_names_dict[a_class] = \
                    build_shapes_name_for_class_uri(class_uri=a_class,
                                                    shapes_namespace=self._class_profiler._shapes_namespace,
                                                    shape_names_dict=self._shape_names_dict,
                                                    namespace_prefix_dict=self._namespaces_dict)

    def has_shape_annotated_features(self, shape_label):
        if shape_label not in self._c_shapes_dict:
            return False
        return len(self._c_shapes_dict[shape_label]) > 0

    def _set_annotation_methods(self):
        if self._examples_mode is None:
            self.annotate_triple_features = self._annotate_triple_features_no_examples
        else:
            self.annotate_triple_features = self._annotate_triple_features_with_examples


    def _annotate_triple_features_with_examples(self, a_triple):
        self._annotate_target_subject(a_triple)
        self._annotate_example_no_inverse(a_triple=a_triple)

    def _annotate_triple_features_no_examples(self, a_triple):
        self._annotate_target_subject(a_triple)

    def _annotate_example_no_inverse(self, a_triple):
        for a_class_key in self._i_dict[str(a_triple[_S])][POS_CLASSES]:
            if not self._shape_feature_examples.has_constraint_example(shape_id=a_class_key,
                                                                       prop_id=str(a_triple[_P])):
                self._shape_feature_examples.set_constraint_example(shape_id=a_class_key,
                                                                    prop_id=str(a_triple[_P]),
                                                                    example=str(a_triple[_O]))








import unittest
from shexer.shaper import Shaper
from test.const import BASE_FILES, default_namespaces
from test.t_utils import no_sharp_in_shepe_names
import os.path as pth

from shexer.consts import TURTLE



_BASE_DIR = BASE_FILES + "no_sharp" + pth.sep  # We just need something with another instantiation property


class TestNoSharpInShapeNames(unittest.TestCase):

    def test_all_classes_no_sharps(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "sharp_chances.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(no_sharp_in_shepe_names(str_result))


import unittest
from shexer.shaper import Shaper
from test.const import BASE_FILES, default_namespaces, BASE_FILES_GENERAL
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth
from shexer.consts import NT, TURTLE

_BASE_DIR = BASE_FILES + "graph_list_of_files_input" + pth.sep


class TestGraphListOfFilesInput(unittest.TestCase):

    def test_one_turtle(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        list_of_url_input=[
                            "https://raw.githubusercontent.com/DaniFdezAlvarez/shexerp3/develop/test/t_files/t_graph_1.ttl"],
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=BASE_FILES_GENERAL + "g1_all_classes_no_comments.shex",
                                                      str_target=str_result))

    def test_one_nt(self):  # Should be nt
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        list_of_url_input=[
                            "https://raw.githubusercontent.com/DaniFdezAlvarez/shexerp3/develop/test/t_files/t_graph_1.nt"],
                        namespaces_dict=default_namespaces(),
                        input_format=NT,
                        all_classes_mode=False,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=BASE_FILES_GENERAL + "g1_all_classes_no_comments.shex",
                                                      str_target=str_result))

    def test_several_nt(self):  # Should be nt
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        list_of_url_input=[
                            "https://raw.githubusercontent.com/DaniFdezAlvarez/shexerp3/develop/test/t_files/graph_list_of_files_input/g1_p1.nt",
                            "https://raw.githubusercontent.com/DaniFdezAlvarez/shexerp3/develop/test/t_files/graph_list_of_files_input/g1_p2.nt"],
                        namespaces_dict=default_namespaces(),
                        input_format=NT,
                        all_classes_mode=False,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=BASE_FILES_GENERAL + "g1_all_classes_no_comments.shex",
                                                      str_target=str_result))

    def test_several_turtle(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        list_of_url_input=[
                            "https://raw.githubusercontent.com/DaniFdezAlvarez/shexerp3/develop/test/t_files/graph_list_of_files_input/g1_p1.ttl",
                            "https://raw.githubusercontent.com/DaniFdezAlvarez/shexerp3/develop/test/t_files/graph_list_of_files_input/g1_p2.ttl"],
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=BASE_FILES_GENERAL + "g1_all_classes_no_comments.shex",
                                                      str_target=str_result))



from shexer.io.shex.formater.statement_serializers.frequency_strategy.base_frequency_strategy import BaseFrequencyStrategy
from shexer.utils.uri import prefixize_uri_if_possible, add_corners_if_it_is_an_uri

from shexer.io.shex.formater.consts import SPACES_GAP_BETWEEN_TOKENS, ANNOTATION_BEGIN

_FREQUENCY_PATTERN = "{:.3f}"

class RatioFreqSerializer(BaseFrequencyStrategy):

    def __init__(self, frequency_ratio_property, frequency_absolute_property, namespaces_dict, decimals=-1):
        """

        :param decimals: it indicates the number of decimals to use to express ratios.
                        When a negative number is provided, decimals won't be controlled
        """
        super().__init__(frequency_ratio_property, frequency_absolute_property, namespaces_dict)
        self._decimals = decimals
        if decimals < 0:
            self.serialize_frequency = self._serialize_freq_unbounded
        elif decimals == 0:
            self.serialize_frequency = self._serialize_freq_int
        else:
            self.serialize_frequency = self._serialize_freq_decimals
            _FREQUENCY_PATTERN = "{:."+ str(decimals) + "f}"


    def serialize_frequency(self, statement):
        raise NotImplementedError("This function will be initialized with a callback during the __init__")


    def _serialize_freq_unbounded(self, statement):
        return str(statement.probability * 100) + " %"

    def _serialize_freq_decimals(self, statement):
        pattern = "{:." + str(self._decimals) +"f} %"
        return pattern.format(statement.probability*100)

    def _serialize_freq_int(self, statement):
        return str(int(statement.probability * 100)) + " %"

    def annotations_for_frequency(self, statement):

        freq_annotation = SPACES_GAP_BETWEEN_TOKENS.join((ANNOTATION_BEGIN,
                                                          add_corners_if_it_is_an_uri(prefixize_uri_if_possible(target_uri=self._frequency_ratio_property,
                                                                                    namespaces_prefix_dict=self._namespaces_dict,
                                                                                    corners=False)),
                                                          self._format_frequency(statement.probability)
                                                          ))
        return [freq_annotation]

    def _format_frequency(self, frequency_raw_number):
        return _FREQUENCY_PATTERN.format(frequency_raw_number)